/* $Id: sfskeyreg.C,v 1.37 2002/08/21 13:54:45 max Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfskey.h"
#include "srp.h"

struct keyregstate {
  bool nokbdnoise;
  bool nosrp;
  u_int nbits;
  str keyname;
  str user;
  str host;
  str realm;
  u_int32_t uid;
  sfskey k;
  sfssrp_parms srpparms;
  sfs_loginres res;
  bool srpparms_set;
  ptr<aclnt> c;
  ptr<sfscon> sc;
  int ntries;
  AUTH *auth;
  sfs_autharg2 arg;
  bool chkflag;
  bool rereg;
  sfsauth_userinfo uinfo;

  keyregstate () : nokbdnoise (false), nosrp (false), nbits (0), 
		   srpparms_set (false), ntries (0), auth (NULL),
		   rereg (false) {}

  void setuser () {
    sfs_aid aid;
    if ((uid = getuid ()) < 0)
      fatal << "Cannot find user ID\n";
    if (uid == 0 && !user) {
      if ((aid = myaid ())) {
	struct passwd *pw;
	uid = aid;
	if (!(pw = getpwuid (uid))) 
	  fatal << "Cannot find pwent for uid " << uid << "\n";
	user = pw->pw_name;
	free (pw);
      } else {
	fatal << "Must supply a username with -u when running as root\n";
      }
    } else if (!user && !(user = myusername ())) {
      fatal << "Cannot find username\n";
    }
  }

  void chk () {
    while (chkflag)
      acheck ();
  }

  void doreg () { 
    setuser ();
    chkflag = true;
    sfs_connect_path ("-", SFS_AUTHSERV, 
		      wrap (this, &keyregstate::getcon) ,
		      true, false);
    chk ();

    query ();
    authenticate ();
    authenticated ();
    getkey ();
    setsrpparms();
    sendrpc ();
    success ();

  }
  
  void getcon (ptr<sfscon> s, str err) {
    if (!s)
      fatal << "Cannot connect to auth server on localhost: " << err << "\n";
    sc = s;
    sfs_hostinfo_upgrade (sc->servinfo);
    host = sfs_get_hostinfo (sc->servinfo).hostname;
    chkflag = false;
    return ;
  }

  void authenticate () {
    c = aclnt::alloc (sc->x, sfs_program_1);
    sfs_loginarg larg;
    arg.set_type (SFS_UNIXPWAUTH);
    arg.pwauth->req.type = SFS_SIGNED_AUTHREQ;
    arg.pwauth->req.authid = sc->authid;
    arg.pwauth->req.user = user; 

    ntries = 0;
    bool goodlogin = false;
    while (ntries < 3) {
      arg.pwauth->req.seqno = ntries;
      if (uid) 
	arg.pwauth->password = getpwd ("  UNIX password: ");
      else 
	arg.pwauth->password = "";
      larg.seqno = ntries;
      if (!xdr2bytes (larg.certificate, arg)) 
	fatal << "Cannot marshall login arguments.\n";
      if (clnt_stat err = c->scall (SFSPROC_LOGIN, &larg, &res))
	fatal << "Cannot attempt UNIX password authentication (" << err 
	      << ")\n";
      if (res.status == SFSLOGIN_OK) {
	auth = authuint_create (*res.authno);
	goodlogin = true;
	break;
      }
      ntries ++;
    }
    if (!goodlogin) 
      fatal << "Too many tries.\n";
  }
  
  void authenticated () {
    c = aclnt::alloc (sc->x, sfsauth_prog_2);
    if (!nosrp) {
      sfsauth2_query_arg aqa;
      sfsauth2_query_res aqr;
      aqa.type = SFSAUTH_CERTINFO;
      aqa.key.set_type (SFSAUTH_DBKEY_NULL);
      if (clnt_stat err = c->scall (SFSAUTH2_QUERY, &aqa, &aqr))
	fatal << "authserv: certinfo: " << err << "\n";
      if (aqr.certinfo->name.len () <= 0)
	fatal << "could not get server's name/realm";
      realm = aqr.certinfo->name;
      if (aqr.certinfo->info.status == SFSAUTH_CERT_REALM)
	warnx << "Authserver is in realm " << realm << "\n";
    }
  }

  void getkey () {
    warnx << "Registering new user " << user << "@" << host << ".\n";
    if (!keyname || (!iskeyremote (keyname)
		     && (access (keyname, 0) < 0 && errno == ENOENT))) {
      k.keyname = user << "@" << host;
      if (str err = sfskeygen (&k, nbits, NULL, keyname, false, nokbdnoise))
	fatal << err << "\n";
      if (keyname)
	if (str err = sfskeysave (keyname, &k))
	  warn << err << "\n";
    }
    else if (str err = sfskeyfetch (&k, keyname))
      fatal << err << "\n";

    if (!k.pwd && !nosrp) {
      fatal << "Cannot use SRP on unencrypted key " << keyname << "\n"
	    << "Specify -S flag to disable SRP\n";
    }

  }

  void setsrpparms () {
    if (nosrp || srpparms_set)
      return;
    if (k.srpparms) {
      srpparms = *k.srpparms;
      return;
    }
    if (!get_srp_params (NULL, &srpparms.N, &srpparms.g)) {
      fatal << "Could not find suitable source for SRP parameters\n";
    }
    srpparms_set = true;
  }

  void query () {
    c = aclnt::alloc (sc->x, sfsauth_prog_2);
    sfsauth2_query_arg qa;
    sfsauth2_query_res qr;
    qa.type = SFSAUTH_USER;
    qa.key.set_type (SFSAUTH_DBKEY_NAME);
    *(qa.key.name) = user;
    if (clnt_stat err = c->scall (SFSAUTH2_QUERY, &qa, &qr))
      fatal << "Error querying auth server: " << err << "\n";
    switch (qr.type) {
    case SFSAUTH_ERROR:
      fatal << "User " << user 
	    << " does not have a UNIX account on this system.\n" ;
      uinfo.name = user;
      uinfo.vers = 1;
      uinfo.gid = 0;
      uinfo.id = 0;
      break;
    case SFSAUTH_USER:
      if (qr.userinfo->vers > 0 && !rereg)
	fatal << "Cannot overwrite existing sfsauthd record.\n" 
	      << "Try the -f flag to override.\n";
      uinfo = *(qr.userinfo);
      uinfo.vers ++;
      break;
    default:
      fatal << "Unexpected response to query.\n";
    }
  }

  void sendrpc () {
    sfsauth2_update_arg aua;
    sfsauth2_update_res aur;
    aua.req.type = SFS_UPDATEREQ;
    aua.req.authid = sc->authid;
    aua.req.rec.set_type (SFSAUTH_USER);
    if (!k.key->export_pubkey (&uinfo.pubkey)) 
      fatal << "Cannot set new public key\n";

    if (!nosrp) {
      srp_client srp;
      str srpi = srp.create (srpparms.N, srpparms.g, k.pwd, realm, k.cost);
      if (!srpi)
	fatal << "Failed to create SRP info\n";
      uinfo.pwauth = srpi;
      str epk;
      if (!k.key->export_privkey (&epk, &srp.eksb))
	fatal << "Cannot encrypt public key.\n";
      uinfo.privkey = epk;
    }

    *(aua.req.rec.userinfo) = uinfo;
    str msg = xdr2str (aua.req);
    if (!msg) 
      fatal << "Could not marshall registration request\n";
    aua.newsig.alloc ();
    if (!k.key->sign (aua.newsig, msg))
      fatal << "Could not sign registration request\n";
    if (!auth) 
      fatal << "Can't procede with update without valid AUTHNO\n";
    if (clnt_stat err = c->scall (SFSAUTH2_UPDATE, &aua, &aur, auth))
      fatal << "SFS auth server: " << err << "\n";
    if (!aur.ok) 
      fatal << "Registration failed: " << *(aur.errmsg) << "\n";
  }

  void success () { exit (0); }

};

void
sfskey_reg (int argc, char **argv)
{
  keyregstate *kr = New keyregstate ();

  int ch;
  while ((ch = getopt (argc, argv, "KSfs:b:c:u:")) != -1)
    switch (ch) {
    case 'f':
      kr->rereg = true;
      break;
    case 'K':
      kr->nokbdnoise = true;
      break;
    case 'S':
      kr->nosrp = true;
      break;
    case 's':
      {
	str sp = file2str (optarg);
	if (!sp)
	  fatal ("%s: %m\n", optarg);
	else if (!import_srp_params (sp, &kr->srpparms.N, &kr->srpparms.g))
	  fatal ("%s: cannot parse file contents\n", optarg);
        else
          kr->srpparms_set = true;
	break;
      }
    case 'b':
      if (!convertint (optarg, &kr->nbits))
	usage ();
      break;
    case 'c':
      if (!convertint (optarg, &kr->k.cost))
	usage ();
      break;
    case 'u':
      kr->user = optarg;
      break;
    default:
      usage ();
      break;
    }
  if (optind + 1 < argc)
    usage ();

  if (optind + 1 == argc)
    kr->keyname = argv[optind];
  else if (optind == argc) {
    if (!kr->user) {
      agent_mkdir ();
      kr->keyname = defkey ();
    }
  }
  else
    usage ();

  random_set_seedfile (RANDOM_SEED);
  kr->doreg ();
}

