/* $Id: sfspath.C,v 1.9 2002/08/21 13:54:48 max Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfscrypt.h"
#include "sfsmisc.h"
#include "crypt.h"
#include "parseopt.h"

#include "rxx.h"
static rxx ppv2 ("([a-z0-9]+)@([\\w\\.\\-]+)(%(\\d+))?");


bool
sfs_ascii2hostid (sfs_hash *hid, const char *p)
{
  const u_char *const up = reinterpret_cast<const u_char *> (p);
  if (armor32len (up) != ascii_hostid_len || p[ascii_hostid_len])
    return false;
  if (hid) {
    str raw = dearmor32 (p, ascii_hostid_len);
    memcpy (hid->base (), raw, hid->size ());
  }
  return true;
}

bool
sfsgethost_label (const char *&p)
{
  const char *s = p;
  while ((*s >= 'a' && *s <= 'z')
	 || (*s >= '0' && *s <= '9') || *s == '-')
    s++;
  if (s - p == 0 || s - p > 63)
    return false;
  p = s;
  return true;
}
bool
sfsgethost_dotlabel (const char *&p)
{
  const char *s = p;
  if (*s++ != '.')
    return false;
  if (!sfsgethost_label (s))
    return false;
  p = s;
  return true;
}
str
sfsgethost (const char *&p)
{
  const char *s = p;
  if (!sfsgethost_label (s) || !sfsgethost_dotlabel (s))
    return NULL;
  while (sfsgethost_dotlabel (s))
    ;
  if (s - p > 255)
    return NULL;

  str host (p, s - p);
  p = s;
  return host;
}

bool
sfs_parsepath (str path, str *host, sfs_hash *hostid, u_int16_t *portp,
	       int *versp)
{
  int vers;
  bool ret;
  const char *p = path;
  if (strchr (p, ':')) {
    vers = 1;
    ret =  sfs_parsepath_v1 (path, host, hostid, portp);
  } else {
    vers = 2;
    ret = sfs_parsepath_v2 (path, host, hostid, portp);
  }
  if (versp) *versp = vers;
  return ret;
}

bool
sfs_parsepath_v2 (str &path, str *hostp, sfs_hash *hostidp, u_int16_t *portp)
{
  u_int16_t port = sfs_port;
  const char *cp;
  sfs_hash dummy;
  if (!hostidp) 
    hostidp = &dummy;
  
  if (!ppv2.match (path))
    return false;
  if (!ppv2[1] || !ppv2[1].len () || !sfs_ascii2hostid (hostidp, ppv2[1]))
    return false;
  if (!ppv2[2] || !ppv2[2].len () || !(cp = ppv2[2]) || !sfsgethost (cp))
    return false;
  if (hostp) 
    *hostp = ppv2[2];
  if (ppv2[4] && ppv2[4].len ()) {
    char *endp;
    int64_t atno = strtoi64 (ppv2[4], &endp, 10);
    if (!endp || *endp || atno <= 0 || atno >= 0x10000)
      return false;
    port = atno;
  }
  if (portp) *portp = port;
  return true;
}

bool
sfs_parsepath_v1 (str &path, str *host, sfs_hash *hostid, u_int16_t *portp)
{
  const char *p = path;
  u_int16_t port = sfs_port;

  if (isdigit (*p)) {
    int64_t atno = strtoi64 (path, const_cast<char **> (&p), 10);
    if (p && *p == '@' && atno > 0 && atno < 0x10000) {
      port = atno;
      path = substr (path, p + 1 - path.cstr ());
    }
    p = path;
  }

  if (portp)
    *portp = port;
  if (!sfsgethost (p))
    return false;
  if (host)
    host->setbuf (path, p - path);
  return *p++ == ':' && sfs_ascii2hostid (hostid, p);
}

bool
sfs_mkhostid (sfs_hash *id, const sfs_hostinfo2 &info, int vers)
{
  switch (vers) {
  case 2:
    return sfs_mkhostid_v2 (id, info);
  case 1:
    return sfs_mkhostid_v1 (id, info);
  default:
    warn << "sfs_mkhostid: unrecognized version number\n";
    return false;
  }
}

bool
sfs_mkhostid_v2 (sfs_hash *id, const sfs_hostinfo2 &info)
{
  const char *p = info.hostname;
  if (info.type != SFS_HOSTINFO || !sfsgethost (p) || *p) {
    bzero (id->base (), id->size ());
    return false;
  }
      
  // Compute SHA-1( SHA-1(k), k )
  sha1ctx sha;
  sfs_pubkey2_hash k;
  k.type = SFS_PUBKEY2_HASH;
  k.pubkey = info.pubkey;
  str s = xdr2str (k);
  u_int8_t h[sha1::hashsize];
  if (s.len () <= 0) {
    warn << "sfs_mkhostid (" << info.hostname << "): XDR failed!\n";
    bzero (id->base (), id->size ());
    return false;
  }
  sha.update (s.cstr (), s.len ());
  sha.final (&h);
  sha.reset ();
  sha.update (h, sha1::hashsize);
  sha.update (s.cstr (), s.len ());
  sha.final (id);
  return true;
}

bool 
sfs_mkhostid (sfs_hash *id, const sfs_hostinfo &info)
{
  const char *p = info.hostname;
  if (info.type != SFS_HOSTINFO || !sfsgethost (p) || *p) {
    bzero (id->base (), id->size ());
    return false;
  }
  
  xdrsuio x;
  if (!xdr_sfs_hostinfo (&x, const_cast<sfs_hostinfo *> (&info))
      || !xdr_sfs_hostinfo (&x, const_cast<sfs_hostinfo *> (&info))) {
    warn << "sfs_mkhostid (" << info.hostname << ", "
	 << hexdump (id->base (), id->size ()) << "): XDR failed!\n";
    bzero (id->base (), id->size ());
    return false;
  }
  sha1_hashv (id->base (), x.iov (), x.iovcnt ());
  return true;
}

bool
sfs_mkhostid_v1 (sfs_hash *id, const sfs_hostinfo2 &v2)
{
  sfs_hostinfo info;
  info.type = v2.type;
  info.hostname = v2.hostname;
  if (v2.pubkey.type != SFS_RABIN)
    return false;
  info.pubkey = *(v2.pubkey.rabin);
  return sfs_mkhostid (id, info);
}

str
sfs_hostinfo2path (const sfs_hostinfo &info)
{
  sfs_hash hostid;
  if (!sfs_mkhostid (&hostid, info))
    return NULL;
  strbuf b;
  b << info.hostname << ":" << armor32 (&hostid, sizeof (hostid)) ;
  return b;
}

str
sfs_hostinfo2path (const sfs_hostinfo2 &info)
{
  sfs_hash hostid;
  if (!sfs_mkhostid (&hostid, info)) 
    return NULL;
  strbuf b;
  b << armor32 (&hostid, sizeof (hostid))  << "@" << info.hostname;
  if (info.port && info.port != sfs_port)
    b << "%" << info.port;
  return b;
}

bool
sfs_ckhostid (const sfs_hash *id, const sfs_hostinfo2 &info, int vers)
{
  sfs_hash rid;
  return sfs_mkhostid (&rid, info, vers) && !memcmp (&rid, id, sizeof (rid));
}

bool
sfs_hostinfo_assert (const sfs_servinfo &sinfo)
{
  return (sinfo.sivers == 7);
}

bool
sfs_hostinfo_server_assert (const sfs_servinfo &sinfo)
{
  return (sinfo.sivers == 7 && sinfo.cr7->release == 7);
}

int
sfs_authd_vers (const sfs_servinfo &sinfo)
{
  assert (sfs_hostinfo_assert (sinfo));
  return (sinfo.cr7->release == 7 ? 2 : 1);
}

const sfs_hostinfo2 &
sfs_get_hostinfo (const sfs_servinfo &sinfo)
{
  return (sinfo.cr7->host);
}

bool
sfs_hostinfo_upgrade (sfs_servinfo &sinfo)
{
  bool ret = true;
  switch (sinfo.sivers) {
  case 5:
  case 6:
    {
      sfs_servinfo_7 cr7;
      cr7.prog = sinfo.cr5->prog;
      cr7.vers = sinfo.cr5->vers; 
      cr7.host.type = sinfo.cr5->host.type;
      cr7.host.hostname = sinfo.cr5->host.hostname;
      cr7.host.pubkey.set_type (SFS_RABIN);
      cr7.release = sinfo.sivers;
      *(cr7.host.pubkey.rabin) = sinfo.cr5->host.pubkey;
      sinfo.set_sivers (sfs_release);
      *(sinfo.cr7) = cr7;
      break;
    }
  case 7:
    break;
  default:
    warn << "sfs_hostinfo_upgrade: unknown sfs_srvinfo info (release " 
	 << sinfo.sivers / 100 << "." << sinfo.sivers % 100 << ")\n";
    ret = false;
  }
  return ret;
}

bool
sfs_ckpath (str sname, const sfs_hostinfo2 &info)
{
  str location;
  sfs_hash hid;
  int vers;
  u_int16_t port;
  return sfs_parsepath (sname, &location, &hid, &port, &vers)
    && location == info.hostname && sfs_ckhostid (&hid, info, vers);
}

bool
sfs_ckci (const sfs_connectinfo &ci, const sfs_hostinfo2 &info)
{
  if (ci.civers <= 4)
    return sfs_ckhostid (&ci.ci4->hostid, info, 1);
  return sfs_ckpath (ci.ci5->sname, info);
}

bool
sfs_checkrevoke (const sfs_pathrevoke &rev)
{
  if (rev.msg.type != SFS_PATHREVOKE
      || rev.msg.path.type != SFS_HOSTINFO
      || (rev.msg.redirect && rev.msg.redirect->hostinfo.type != SFS_HOSTINFO))
    return false;

  if (!sfscrpt.verify (rev.msg.path.pubkey, rev.sig, xdr2str (rev.msg)))
    return false;

  sfs_hash hostid;
  if (rev.msg.redirect &&
      (!sfs_mkhostid (&hostid, rev.msg.redirect->hostinfo)
       || implicit_cast<time_t> (rev.msg.redirect->expire) >= timenow))
    return false;

  return true;
}
