
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "gcc_compat.h"
#include <dhashgateway_prot.h>
#include "dhash_common.h"
#include "aquasingle.h"
#include "dbfe.h"
#include "arpc.h"
#include "verify.h"
#include "merkle_misc.h"

extern ptr<dbfe> cache_db;

ptr<dhash_retrieve_res>
block_to_res (ptr<dbrec> val, dhash_ctype ctype)
{
  ptr<dhash_retrieve_res> res = New refcounted<dhash_retrieve_res> (DHASH_OK);
  int n = val->len;
  res->resok->ctype = ctype;
  res->resok->len = n;
  res->resok->hops = 0;
  res->resok->errors = 0;
  res->resok->retries = 0;
  res->resok->path.setsize(0);
  res->resok->block.setsize(n);
  memcpy (res->resok->block.base (), val->value, n);
  return res;
}

bool
is_keyhash_stale (ref<dbrec> prev, ref<dbrec> d)
{
  long v0 = keyhash_version (prev);
  long v1 = keyhash_version (d);
  if (v0 >= v1)
    return true;
  return false;
}

aquasinglesrv::aquasinglesrv (ptr<axprt_stream> x)
{
  clntsrv = asrv::alloc (x, dhashgateway_program_1,
	                 wrap (mkref (this), &aquasinglesrv::dispatch));
}

aquasinglesrv::~aquasinglesrv ()
{
}

void
aquasinglesrv::insert_to_localcache
  (chordID id, char* block, int32_t len, dhash_ctype ctype)
{
  // insert into DB
  ref<dbrec> k = id2dbrec(id);
  ref<dbrec> d = New refcounted<dbrec> (block, len);
  ptr<dbrec> prev;

  if (!verify (id, ctype, block, len)) {
    warn  << "aquasingle: cannot verify (" << len << ") " << id << " bytes\n";
    assert(0);
  }

  switch(ctype) {
  case DHASH_CONTENTHASH:
    if (!cache_db->lookup (k)) {
      cache_db->insert (k, d);
      warn << "db: " << ctype << " " << id << " " << len << "\n";
    } else {
      warn << "db: " << ctype << " " << id << " exists\n";
    }
    break;

  case DHASH_KEYHASH: 
    prev = cache_db->lookup(k);
    if (prev) {
      if (is_keyhash_stale(prev, d)) {
	warn << "db: " << ctype << " " << id << " " << len << " (stale)\n";
	break;
      }
      else {
	warn << "db: " << ctype << " " << id << " " << len << " (rw)\n";
	cache_db->del(k);
      }
    }
    else
      warn << "db: " << ctype << " " << id << " " << len << "\n";
    cache_db->insert (k, d);
    break;

  case DHASH_NOAUTH:
    prev = cache_db->lookup(k);
    if (prev)
      cache_db->del(k);
    cache_db->insert(k, d);
    warn << "db: " << ctype << " " << id << " " << len << " (noauth)\n";
    break;

  case DHASH_APPEND:
  case DHASH_UNKNOWN:
  default:
    warn << "aquasingle can't handle inserting ctype: " << ctype << "\n";
  }

  return;
}

void
aquasinglesrv::dispatch (svccb *sbp)
{
  if (!sbp) {
    // setting clntsrv to 0 removes the last reference to this gateway
    // object, stored in the asrv object's callback.
    clntsrv = 0;
    return;
  }

  switch (sbp->proc ()) {
  case DHASHPROC_NULL:
    sbp->reply (NULL);
    return;
    
  case DHASHPROC_INSERT:
    {
      dhash_insert_arg *arg = sbp->TEMPLATE getarg<dhash_insert_arg> ();
      insert_to_localcache
        (arg->blockID, arg->block.base(), arg->block.size(), arg->ctype);
      ptr<dhash_insert_res> res = New refcounted<dhash_insert_res> (DHASH_OK);
      sbp->reply (res);
    }
    break;
    
  case DHASHPROC_RETRIEVEPK:
    {
      dhashgw_retrieve_pk_arg *arg =
	sbp->TEMPLATE getarg<dhashgw_retrieve_pk_arg> ();
      ptr<dbrec> cache_ret = cache_db->lookup (id2dbrec (arg->key));
      if (cache_ret) {
	ptr<dhash_retrieve_res> res = block_to_res (cache_ret, DHASH_KEYHASH);
	sbp->reply (res);
	return;
      }
      else {
	dhash_retrieve_res res (DHASH_NOENT);
	sbp->reply (&res);
      }
    }
    break;

  case DHASHPROC_RETRIEVE:
    {
      dhash_retrieve_arg *arg = sbp->TEMPLATE getarg<dhash_retrieve_arg> ();
      ptr<dbrec> cache_ret = cache_db->lookup (id2dbrec (arg->blockID));
      if (cache_ret) {
	ptr<dhash_retrieve_res> res = block_to_res (cache_ret, arg->ctype);
	sbp->reply (res);
	return;
      }
      else {
	dhash_retrieve_res res (DHASH_NOENT);
	sbp->reply (&res);
      }
    }
    break;

  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

