/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "pastdbfe.h"
#include "btree.h"

void
btree_node::serialize (strbuf &b) const
{
  commit.serialize (b);
  prev.serialize (b);
  suint (b, level);
  sstr (b, proposed_name);
  suint (b, next.size ());
  for (unsigned i=0; i<next.size (); i++)
    next [i].serialize (b);
}

bool
btree_node::parse (xin &x, btree_node &n)
{
  unsigned k;
  if (!commit_info::parse (&x, n.commit) ||
      !commit_info::parse (&x, n.prev) ||
      !x.ruint (n.level) ||
      !x.rstr (n.proposed_name) ||
      !x.ruint (k))
    return false;
  for (unsigned i=0; i<k; i++) {
    commit_info a;
    if (!commit_info::parse (&x, a))
      return false;
    n.next.push_back (a);
  }
  return true;
}

btree_srv::btree_srv (ptr<pastdbfe> db)
  : _db (db)
{
}

static bool
dbrec2node (ptr<dbrec> d, btree_node &n)
{
  str s (d->value, d->len);
  xin x (s);
  return btree_node::parse (x, n);
}

static ptr<dbrec>
node2dbrec (const btree_node &n)
{
  strbuf b;
  n.serialize (b);
  str s = b;
  return New refcounted<dbrec> (s.cstr (), s.len ());
}

void
btree_srv::add_node (bigint c, btree_node &n)
{
  ptr<dbrec> k = bigint2dbrec (c);
  if (_db->lookup (k)) {
    warn << armor64A (c) << " already exists in branch tree, replace!\n";
    _db->remove (k);
  }

  ptr<dbrec> p = bigint2dbrec (n.prev.rptr);
  ptr<dbrec> d = _db->lookup (p);
  if (d) {
    btree_node o;
    if (!dbrec2node (d, o)) { assert (0); }
    o.next.push_back (n.commit);
    n.level = o.level+1;
    _db->remove (p);
    
    ptr<dbrec> d0 = node2dbrec (o);
    _db->insert (p, d0);
    
    ptr<dbrec> d1 = node2dbrec (n);
    _db->insert (k, d1);
  }
  else {
    assert (n.prev.rptr == 0);
    n.level = 0;
    ptr<dbrec> d1 = node2dbrec (n);
    _db->insert (k, d1);
  }

  add_tag_aux (c, n.proposed_name, false);
}

void
btree_srv::add_tag_aux (bigint c, str tag, bool tagged)
{
  unsigned n = 1;
  vec<bigint> commits;
  commits.push_back (c);
  
  ptr<dbrec> tag_k = str2dbrec (tag);
  ptr<dbrec> tag_v = _db->lookup (tag_k);
  if (tag_v) {
    _db->remove (tag_k);
    str s = dbrec2str (tag_v);
    xin x (s);
    unsigned i;
    x.ruint (i);
    n += i;
    for (unsigned j=0; j<i; j++) {
      bigint k;
      x.rbigint (k);
      commits.push_back (k);
    }
  }
  strbuf b;
  suint (b, n);
  for (unsigned i=0; i<n; i++)
    sbigint (b, commits [i]);
  tag_v = str2dbrec (b);
  _db->insert (tag_k, tag_v);

  if (tagged) {
    ptr<dbrec> k = str2dbrec ("tags");
    ptr<dbrec> v = _db->lookup (k);
    str x;
    if (v) {
      _db->remove (k);
      x = dbrec2str (v);
    }
    else
      x = "";
    strbuf b;
    sstr (b, tag);
    b << x;
    ptr<dbrec> d = str2dbrec (b);
    _db->insert (k, d);
  }
}

str
btree_srv::get_tags ()
{
  ptr<dbrec> k = str2dbrec ("tags");
  ptr<dbrec> v = _db->lookup (k);
  str x;
  if (v)
    x = dbrec2str (v);
  return x;
}

void
btree_srv::add_tag (bigint c, str tag)
{
  add_tag_aux (c, tag, true);
}

bool
btree_srv::lookup_by_commit (bigint c, btree_node &n)
{
  ptr<dbrec> k = bigint2dbrec (c);
  ptr<dbrec> d = _db->lookup (k);
  if (d){
    return dbrec2node (d, n);
  }

  return false;
}

bool
btree_srv::lookup_by_tag (str t, bigint &c)
{
  if (!t || t == "")
    return false;

  ptr<dbrec> k = str2dbrec (t);
  ptr<dbrec> d = _db->lookup (k);

  if (d) {
    str s = dbrec2str (d);
    xin x (s);
    unsigned i;
    x.ruint (i);
    if (i == 1) {
      x.rbigint (c);
      return true;
    }
  }

  return false;
}

bool
btree_srv::find_join (bigint c0, bigint c1, bigint &cj)
{
  if (c0 == c1) {
    cj = c0;
    return true;
  }

  btree_node x0, x1;
  if (!lookup_by_commit (c0, x0) || !lookup_by_commit (c1, x1))
    return false;
  assert (x0.level > 0 || x1.level > 0);

  if (x0.level < x1.level)
    return find_join (c0, x1.prev.rptr, cj);
  else
    return find_join (x0.prev.rptr, c1, cj);
}

bool
btree_srv::find_leaves (bigint c, vec<btree_node>& v)
{
  vec<bigint> todo;
  v.clear ();
  todo.push_back (c);

  while (todo.size ()) {
    bigint n = todo.pop_front ();
    btree_node x;
    if (!lookup_by_commit (n, x))
      return false;
    if (x.next.size () == 0)
      v.push_back (x);
    else {
      for (unsigned i=0; i<x.next.size (); i++) {
	// warn << "next is " << armor64A (x.next [i]);
	todo.push_back (x.next [i].rptr);
      }
    }
  }
  return true;
}

bool
btree_srv::get_sole_descendant (bigint cr, bigint &head)
{
  // Look for a sole descendant branch head of <cr>.  If a sole descendant
  // exists, then return true, and set <head> equal to the branch head's commit
  // record address.

  bigint current = cr;
  btree_node node;

  while( lookup_by_commit (current, node) ) {
    if (node.next.size() == 0) {
      //warn << "cr " << armor64A(node.commit.rptr) << " is head\n";
      head = node.commit.rptr;
      return true;
    } else if (node.next.size() == 1) {
      //warn << "cr " << armor64A(node.commit.rptr) << " has one child\n";
      current = node.next[0].rptr;
      for(unsigned i=0; i<node.next.size(); i++)
	warn << "  " << armor64A (node.next[i].rptr) << "\n";
    } else {
      //warn << "cr " << armor64A(node.commit.rptr) << " has multiple children\n";
      //for(unsigned i=0; i<node.next.size(); i++)
      //warn << "  " << armor64A (node.next[i].rptr) << "\n";
      return false;
    }
    node.next.clear();
  }
  return false;
}

void
btree_srv::print_node(bigint n) {
  btree_node node;
  if (!lookup_by_commit (n, node))
    return;

  for(unsigned i=0; i<node.next.size(); i++)
    warn << "    " << armor64A(node.next[i].rptr) << "\n";
  
  // recurse
  for(unsigned i=0; i<node.next.size(); i++) {
    print_node(node.next[i].rptr);
  }
}

void
btree_srv::print(bigint start) 
{
  btree_node node;
  bigint current = start;
  
  // find head
  while (lookup_by_commit (current, node))
    current = node.prev.rptr;

  // traverse tree
  print_node(node.commit.rptr);
  return;
}
