/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "gcc_compat.h"
#include "stdinc.h"
#include "pastdbfe.h"
#include "build.h"
#include "snapshot.h"
#include "cfsw_record.h"
#include "btree.h"
#include "replicaimpl.h"

#define treeop_fail(a) \
  treeop_failures++; \
  if (!error || a == CFSRWERR_VERCONFL) error = a

extern ptr<pastdbfe> create_inmem_db ();

ptr<snapshot>
bob::new_snapshot (ptr<snapshot> prev_s)
{
  ptr<snapshot> sobj = New refcounted<snapshot> ();

  if (!prev_s) {
    sobj->thisinfo.rootdir = 0;
    if (!(sobj->fmap_db = create_inmem_db ())) {
      warn << "cannot create fmap database\n";
      return (ptr<snapshot>) 0L;
    }
    return sobj;
  }

  sobj->thisinfo.rootdir = prev_s->thisinfo.rootdir;
  sobj->thisinfo.prevaddr = prev_s->thisaddr;

  if (!(sobj->fmap_db = create_inmem_db ())) {
    warn << "cannot create fmap database\n";
    return (ptr<snapshot>) 0L;
  }

  ptr<pdbEnumeration> it = prev_s->fmap_db->enumerate();
  ptr<dbrec> k;
  while (k = it->nextElement()) {
    ptr<dbrec> v = prev_s->fmap_db->lookup (k);
    bigint chash = dbrec2bigint (v);
    ptr<dbrec> d = bigint2dbrec (chash);
    if (sobj->fmap_db->insert (k, d))
      return (ptr<snapshot>) 0L;
  }
  return sobj;
}

//
// bob class builds a new snapshot object from a set of
// deltas and a base snapshot object
//

bob::bob (replica_impl *r, bigint va, ptr<snapshot> base,
          const vec<deltap>& deltas, crec cr,
          commit_info commit, cbs_t scb)
  : _replica (r), viewaddr (va), deltas (deltas),
    cr (cr), commit (commit), scb (scb)
{
  storing_tree = false;
  outstanding_treeops = 0;
  treeop_failures = 0;
  error = CFSRW_OK;
  history_update_rpcs = 0;
  ndeltas = 0;
  insert_failures = 0;
  outstanding_inserts = 0;
  inserts = 0;
  nchecks = 0;

  sobj = new_snapshot (base);
  if (!sobj) {
    warn << "cannot create new snapshot object\n";
    scb ((ptr<snapshot>) 0, CFSRWERR_ERROR);
    return;
  }
  
  sobj->thisinfo.branch = commit;
  sobj->thisinfo.flags = cr->flags ();
  sobj->thisinfo.supersede = cr->supersede ();

  // if user specified a branch tag, set it; otherwise, inherit the
  // user-defined branch tag from previous snapshot.
  if (cr->btag () != "")
    sobj->thisinfo.btag = cr->btag ();
  else if (base)
    sobj->thisinfo.btag = base->thisinfo.btag;
  else
    sobj->thisinfo.btag = "init";

  next ();
}

void
bob::insert_history (cfs_history *history, cfs_version &v)
{
  if (v.revision == 1) { // first one to be evicted
    history->chain.v = v;
    for (int i=0; i<CFS_VCN; i++) {
      history->chain.ptrs [i] = 0;
      history->chain.paddr [i] = 0;
    }
    return;
  }

  // put history->chain into DHash; write v into history->chain. also,
  // based on the previous pointers in history->chain, compute a new
  // set of pointers in history->chain. pointers look like skip-list
  // pointers.

  cfs_vchain c = history->chain;
  history->chain.v = v;
  int i, j;
  for (j=0, i=1; j < CFS_VCN && i < CFS_VCN; i++) {
    unsigned a = (1 << i);
    if (a < v.revision)
      a = v.revision - a;
    else
      a = 0;
    if (j < CFS_VCN) {
      while (j+1 < CFS_VCN) {
	if (c.ptrs [j+1] < a)
	  break;
	j++;
      }
    }
    history->chain.ptrs [i] = c.ptrs [j];
    history->chain.paddr [i] = c.paddr [j];
  }

  history->chain.ptrs [0] = c.v.revision;
  history->chain.paddr [0] = write_chain (c);
  return;
}

bigint
bob::write_chain (cfs_vchain &c)
{
  strbuf sb;
  sint (sb, c.v.revision);
  sbigint (sb, c.v.logid);
  sbigint (sb, c.v.commit);
  sbigint (sb, c.v.saddr);
  for (unsigned i=0; i<CFS_VCN; i++) {
    sint (sb, c.ptrs [i]);
    sbigint (sb, c.paddr [i]);
  }

  str s = sb;
  inserts++;
  return nsobj->insert_local (s.cstr (), s.len ());
}

int
bob::incr_revision (cfs_attr *attr, cfs_history *history, xrp &x, bool mode)
{
  bigint old_commit = mode ? attr->mode_commit : attr->commit;

  if (old_commit != commit.rptr) { // new version
    attr->revision++;
    if (mode)
      attr->mode_commit = commit.rptr;
    else
      attr->commit = commit.rptr;

    // put the revision in version history. if snapshot addr is
    // available for previous revision (that is, last commit was
    // written in the previous snapshot), save the snapshot addr too.

    unsigned n = history->versions.size ();
    rpc_vec<cfs_version, CFS_NVH> nv = history->versions;
    if (n < CFS_NVH)
      history->versions.setsize (n+1);
    else
      history->versions.setsize (n);
    
    if (n == CFS_NVH)
      insert_history (history, nv [n-1]);

    for (unsigned j=0; j<n; j++) {
      if (1+j < history->versions.size ())
	history->versions [1+j] = nv [j];
    }

    history->versions [0].revision = attr->revision;
    history->versions [0].logid = cr->head ();
    history->versions [0].commit = commit.rptr;
    history->versions [0].saddr = 0;
    history->versions [0].timestamp = cr->time ();

    if (history->versions.size () > 1 &&
	history->versions [1].revision == attr->revision-1)
      history->versions [1].saddr = sobj->thisinfo.prevaddr;
  }
  return 0;
}

void 
bob::next ()
{
  if (outstanding_treeops == 0)
    timecb (tsnow, wrap (mkref (this), &bob::build_tree));
}

void
bob::build_tree ()
{
  if (outstanding_treeops > 0)
    return;

  // build a filesystem tree one operation at a time; avoid handling
  // deltas concurrenly, so we don't have to use locks

  assert (sobj->fmap_db);

  if (!deltas.empty() && !treeop_failures) {
    ndeltas++;
    xrp x = New refcounted<xr>;
    x->record = deltas.pop_front ();

    if (inode_delta *r =
	x->record->TEMPLATE record<inode_delta> ()) {
      ptr<dbrec> q = bigint2dbrec (r->object ());
      ptr<dbrec> v = sobj->fmap_db->lookup (q);
      if (v || sobj->active_snodes [r->object ()]) {
	warn << "cannot create " << armor64A (r->object ())
	     << ", already exists\n";
	treeop_fail (CFSRWERR_ERROR);
      }
      else {
	if (sobj->thisinfo.rootdir == 0 && r->object_type () != CFSW_DIR) {
          warn << "first i-node in a repository must be a directory.\n";
	  treeop_fail (CFSRWERR_ERROR);
	}
	else {
	  if (sobj->thisinfo.rootdir == 0)
	    sobj->thisinfo.rootdir = r->object ();
	  ptr<cfs_data> sdata = New refcounted<cfs_data> (CFS_INODE);
	  create_inode (x, sdata->inode);
	  sobj->update_index (r->object (), sdata);
	}
      }
      if (outstanding_treeops == 0)
        next ();
    }
    else if (sobj->thisinfo.rootdir == 0) {
      warn << "cannot process delta w/o creating a root directory.\n";
      treeop_fail (CFSRWERR_ERROR);
      next ();
    }
    else if (x->record->rtype () == cfsw_delta::T_LINK)
      do_link (x);
    else if (x->record->rtype () == cfsw_delta::T_UNLINK)
      do_unlink (x, 0);
    else if (x->record->rtype () == cfsw_delta::T_SYMLINK)
      do_write (x);
    else if (x->record->rtype () == cfsw_delta::T_WRITE)
      do_write (x);
    else if (x->record->rtype () == cfsw_delta::T_PATCH)
      do_patch (x);
    else if (x->record->rtype () == cfsw_delta::T_ATTR)
      setattr (x);
    else if (x->record->rtype () == cfsw_delta::T_RENAME)
      do_rename (x);
    else if (x->record->rtype () == cfsw_delta::T_TAG) {
      tag_delta *r = x->record->TEMPLATE record<tag_delta> ();
      btree_node n;
      if (r->commit () == 0) { // tag self
	_replica->btree ()->add_tag (commit.rptr, r->tag ());
	next ();
	return;
      }
      else if (_replica->btree ()->lookup_by_commit (r->commit (), n)) {
        // commit exists
	_replica->btree ()->add_tag (r->commit (), r->tag ());
	next ();
	return;
      }
      else {
        warn << "commit does not exist, cannot tag\n";
	treeop_fail (CFSRWERR_ERROR);
	next ();
	return;
      }
    }
    else
      next ();
  }
  else if (!treeop_failures) {
    assert (outstanding_treeops == 0);
    assert (nchecks == 0 || !storing_tree);
    if (nchecks == 0 && !storing_tree) {
      storing_tree = true;
      store_tree ();
    }
  }
  else {
    warn << "cannot build snapshot\n";
    scb ((ptr<snapshot>) 0, error);
  }
}

void 
bob::create_inode (xrp x, cfs_inode *inode)
{
  inode_delta *r = x->record->TEMPLATE record<inode_delta>();

  switch (r->object_type ()) {
  case CFSW_DIR:    // ignore dir opaque case
  case CFSW_REG:
    {
      cfs_ftype t = (r->object_type () == CFSW_DIR) ? CFSDIR : CFSREG;
      inode->set_type (t); 
      rpc_clear (*inode->reg);
      inode->reg->handle = r->object ();
      inode->reg->attr.nlink = 0;
      
      init_attr (inode->reg);
      init_history (&inode->reg->history);

      if (incr_revision (&inode->reg->attr, &inode->reg->history, x)) {
	treeop_fail (CFSRWERR_VERCONFL);
      }
      else {
        set_attr (inode->reg, r->attr ());
        inode->reg->direct.setsize (0);
        if (t == CFSREG)
          write_file (x, inode->reg);
      }
    }
    break;
  case CFSW_LNK:
    {
      inode->set_type (CFSLNK);
      rpc_clear (*inode->lnk);
      inode->lnk->handle = r->object ();
      inode->lnk->attr.nlink = 0;
  
      init_attr (inode->lnk);
      init_history (&inode->lnk->history);

      if (incr_revision (&inode->lnk->attr, &inode->lnk->history, x)) {
	treeop_fail (CFSRWERR_VERCONFL);
      }
      else {
        set_attr (inode->lnk, r->attr ());
        // this gets set later
        inode->lnk->dest = "";
      }
    }
    break;
  default:
    warn << "strange inode type, cannot create inode\n";
    treeop_fail (CFSRWERR_ERROR);
  }
}

void 
bob::do_link (xrp x)
{
  outstanding_treeops++;
  dirop_delta *dlr = dirop_record (x->record);
  snode *d = sobj->active_snodes [dlr->dir ()];
  if (!d) {
    sobj->lookup_fmap
      (dlr->dir (), 
       wrap (mkref (this), &bob::do_link_got_dir, x));
  }
  else
    do_link_got_dir (x, d);
}

void
bob::do_link_got_dir (xrp x, snode *d)
{
  if (!d || !d->ent || d->ent->type != CFS_INODE ||
      d->ent->inode->type != CFSDIR) {
    warn << "link: cannot retrieve snode of directory\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  assert (dirop_record (x->record));
  snode *e = sobj->active_snodes [dirop_record (x->record)->object ()];

  if (!e) {
    sobj->lookup_fmap
      (dirop_record (x->record)->object (),
       wrap (mkref (this), &bob::do_link_got_node, x, d));
  }
  else
    do_link_got_node (x, d, e);
}

void
bob::do_link_got_node (xrp x, snode *d, snode *e)
{
  if (!e || !e->ent || e->ent->type != CFS_INODE) {
    warn << "link: object does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  cfs_inode_reg *dnode = d->ent->inode->reg;    
  bigint dirblk;
  
  if (dnode->direct.size () == 0) {
    // create a new block for this directory
    ptr<cfs_data> dir = New refcounted<cfs_data> (CFS_DIRBLK);
    getrnd (&dirblk, CFS_FHSIZE);
    sobj->update_index (dirblk, dir);
    dnode->direct.setsize (1);
    dnode->direct[0] = dirblk;
    dir->dir->eof = true;
    dir->dir->count = 0;
    add_dirent (x, dnode, dir, e);
    outstanding_treeops--;
    next ();
    return;
  }

  else {
    ptr<cfs_data> free_dir = NULL;
    dirblk = dnode->direct[0];
    snode *dbe = sobj->active_snodes [dirblk];
    if (dbe)
      do_link_got_dirblk (x, d, e, 0, free_dir, dbe);
    else {
      sobj->lookup_fmap
	(dirblk,
	 wrap (mkref (this), &bob::do_link_got_dirblk,
	       x, d, e, 0, free_dir));
    }
  }
}

#define SNAPSHOT_D_NENTRIES 256

void
bob::do_link_got_dirblk 
  (xrp x, snode *d, snode *e, uint32 blknum,
   ptr<cfs_data> free_dir, snode *dbe)
{
  if (!dbe || dbe->ent->type != CFS_DIRBLK) {
    warn << "link: cannot load directory block\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
  
  ptr<cfs_data> dir = dbe->ent;
  cfs_inode_reg *dnode = d->ent->inode->reg;

  if (dir->dir->count < SNAPSHOT_D_NENTRIES && !free_dir)
    free_dir = dir;

  str name = dirop_record (x->record)->name ();
  cfs_dirent* de = dir->dir->entries;
  if (de && (de->name == name)) {
    warn << name << " already exists in directory\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
  else {
    while (de && de->nextentry && (de->nextentry->name != name))
      de = de->nextentry;
    if (!de || !de->nextentry) {
      if (blknum != dnode->direct.size ()-1) {
        blknum++;
	bigint dirblk = dnode->direct[blknum];
        snode *dbe = sobj->active_snodes [dirblk];
        if (dbe)
	  do_link_got_dirblk (x, d, e, blknum, free_dir, dbe);
        else
          sobj->lookup_fmap
	    (dirblk,
	     wrap (mkref (this),
	           &bob::do_link_got_dirblk, 
		   x, d, e, blknum, free_dir));
        return;
      }
    }
    else {
      warn << name << " already exists in directory\n";
      treeop_fail (CFSRWERR_ERROR);
      outstanding_treeops--;
      next ();
      return;
    }
  }

  if (free_dir)
    add_dirent (x, dnode, free_dir, e);
  else
    add_dirent (x, dnode, dir, e);
  outstanding_treeops--;
  next ();
}

void
bob::add_dirent (xrp x, cfs_inode_reg *dnode,
                              ptr<cfs_data> dir, snode *fe)
{
  // directory size is bytes needed to store directory, so we set that
  // before we insert directory block into dhash, not here.

  // after we add an entry to a directory block, we check if the count
  // equals SNAPSHOT_D_NENTRIES. if so, we create a new directory
  // block for the next entry to be added to the directory. we won't
  // create a new directory block if we already created CFS_NDIR
  // number of directory blocks. hence, if the current directory block
  // contains SNAPSHOT_D_NENTRIES number of entries, then we know we
  // hit the directory size limit.
  //
  // CFS_NDIR = 128, so that's 128*256 files per directory.

  if (dir->dir->count == SNAPSHOT_D_NENTRIES) {
    warn << "cannot add file to directory: maximum "
         << (CFS_NDIR * SNAPSHOT_D_NENTRIES) << " files.\n";
    treeop_fail (CFSRWERR_ERROR);
    return;
  }
 
  bigint obj = dirop_record (x->record)->dir ();
  if (incr_revision (&dnode->attr, &dnode->history, x)) {
    treeop_fail (CFSRWERR_VERCONFL);
    return;
  }

  assert (dirop_record (x->record));

  bigint inum = dirop_record (x->record)->object ();

  rpc_ptr<cfs_dirent> direnip;
  direnip.alloc ();
  direnip->fh = inum;
  direnip->name = dirop_record (x->record)->name ();

  if (dir->dir->entries)
    direnip->nextentry = dir->dir->entries;
  dir->dir->entries = direnip;
  dir->dir->count++;
    
  if (dir->dir->count == SNAPSHOT_D_NENTRIES) {
    int blk = dnode->direct.size ();
    if (blk < CFS_NDIR) {
      dir->dir->eof = false;
      // create new directory block
      bigint dirblk;
      ptr<cfs_data> newdir = New refcounted<cfs_data>(CFS_DIRBLK);
      getrnd (&dirblk, CFS_FHSIZE);
      sobj->update_index (dirblk, newdir);
      newdir->dir->eof = true;
      newdir->dir->count = 0;

      rpc_vec<bigint, CFS_NDIR> dv = dnode->direct;
      dnode->direct.setsize (blk+1);
      for (int j=0; j<blk; j++) dnode->direct[j] = dv[j];
      dnode->direct[blk] = dirblk;
    }
  }

  cfs_inode *inode = (fe->ent)->inode;
  if (inode->type == CFSLNK)
    inode->lnk->attr.nlink++;
  else {
    inode->reg->attr.nlink++;
    if (inode->type == CFSDIR) // take into account of ..
      dnode->attr.nlink++;
  }
}

void 
bob::do_write (xrp x)
{
  outstanding_treeops++;
  
  write_delta *w = x->record->TEMPLATE record<write_delta>();
  symlink_delta *s = x->record->TEMPLATE record<symlink_delta>();
  bigint obj = w ? w->file () : s->file ();

  snode *f = sobj->active_snodes [obj];
  if (f)
    do_write_got_node (x, f);
  else
    sobj->lookup_fmap
      (obj, wrap (mkref (this), &bob::do_write_got_node, x));
}

void
bob::do_write_got_node (xrp x, snode *f)
{
  if (!f || f->ent->type != CFS_INODE) {
    warn << "write: file does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  cfs_inode *inode = (f->ent)->inode;
  if (inode->type == CFSLNK) {
    if (inode->lnk->attr.nlink == 0) {
      warn << "cannot write to file with link count 0\n";
      treeop_fail (CFSRWERR_ERROR);
      outstanding_treeops--;
      next ();
      return;
    }
    if (x->record->rtype () != cfsw_delta::T_SYMLINK) {
      warn << "cannot write to symlink\n";
      treeop_fail (CFSRWERR_ERROR);
      outstanding_treeops--;
      next ();
      return;
    }
    do_symlink (x, inode->lnk);
    outstanding_treeops--;
    next ();
  }
  else if (inode->type == CFSREG) {
    if (inode->reg->attr.nlink == 0) {
      warn << "cannot write to file with link count 0\n";
      treeop_fail (CFSRWERR_ERROR);
      outstanding_treeops--;
      next ();
      return;
    }
    if (x->record->rtype () != cfsw_delta::T_WRITE) {
      warn << "cannot create symlink using regular file\n";
      treeop_fail (CFSRWERR_ERROR);
      outstanding_treeops--;
      next ();
      return;
    }
    outstanding_treeops--;
    write_file (x, inode->reg);
    if (outstanding_treeops == 0)
      next ();
  }
  else {
    warn << "write: wrong file type in snapshot inode\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
  }
}

void 
bob::do_symlink (xrp x, cfs_inode_lnk *lnode)
{
  symlink_delta *se = x->record->TEMPLATE record<symlink_delta>();
  if (se->dest ().len () > MAXPATHLEN) {
    warn << "symbolic link target too large\n";
    treeop_fail (CFSRWERR_ERROR);
    return;
  }
  if (incr_revision (&lnode->attr, &lnode->history, x)) {
    treeop_fail (CFSRWERR_VERCONFL);
    return;
  }

  lnode->dest = nfspath3 (se->dest ());
}

void
bob::check_write_fetch (ptr<ns_block> b)
{
  assert (nchecks > 0);
  nchecks--;
  if (!b) {
    warn << "cannot load a block in write delta\n";
    treeop_fail (CFSRWERR_ERROR);
  }
  if (nchecks == 0)
    next ();
}

void
bob::iter_write_cb (bigint n, cfsrw_status status)
{
  if (status) {
    if (status != CFSRWERR_NOMORELOG) {
      warn << "cannot fetch a block in a write delta\n";
      treeop_fail (CFSRWERR_ERROR);
    }
    if (nchecks == 0)
      next ();
  }
  else {
    if (nsobj->check_local (n)) {
      if (nchecks == 0)
        next ();
    }
    else {
      nchecks++;
      nsobj->retrieve_block (n, wrap (mkref (this), &bob::check_write_fetch));
    }
  }
}

extern void iterate_write
(deltap r, callback<void,bigint,cfsrw_status>::ref cb);

void 
bob::write_file (xrp x, cfs_inode_reg *rnode)
{
  write_delta *w = x->record->TEMPLATE record<write_delta> ();
  inode_delta *i = x->record->TEMPLATE record<inode_delta> ();
  attr_delta  *a = x->record->TEMPLATE record<attr_delta> ();

  if (w) {
    rnode->direct.clear ();
    rnode->direct.setsize (w->ndirects ());
    for (unsigned i=0; i < w->ndirects (); i++)
      rnode->direct[i] = w->direct (i);
    rnode->indirect = w->indirect ();
    rnode->double_indirect = w->double_indirect ();
    rnode->triple_indirect = w->triple_indirect ();
    rnode->attr.size = w->size ();

    if (incr_revision (&rnode->attr, &rnode->history, x)) {
      treeop_fail (CFSRWERR_VERCONFL);
      return;
    }
    iterate_write (x->record, wrap (mkref (this), &bob::iter_write_cb));
    return;
  }

  else if (i || a) {
    if ((i && i->attr ()->valid_size) || (a && a->attr ()->valid_size)) {
      uint64 size = i ? i->attr ()->size : a->attr ()->size;
      rnode->attr.size = size;
      if (incr_revision (&rnode->attr, &rnode->history, x)) {
	treeop_fail (CFSRWERR_VERCONFL);
        return;
      }
    }
    return;
  }

  assert (0);
}

static bigint
unlink_parse_dir (deltap e, int phase)
{
  if (e->TEMPLATE record<rename_delta> () && phase == 1)
    return e->TEMPLATE record<rename_delta> ()->from_dir ();
  dirop_delta *dlr = dirop_record (e);
  return dlr->dir ();
}

static str
unlink_parse_name (deltap e, int phase)
{
  if (e->TEMPLATE record<rename_delta> () && phase == 1)
    return e->TEMPLATE record<rename_delta> ()->from_name ();
  dirop_delta *dlr = dirop_record (e);
  return dlr->name ();
}

void 
bob::do_unlink (xrp x, int phase)
{
  outstanding_treeops++;
  bigint dir = unlink_parse_dir (x->record, phase);
  snode *d = sobj->active_snodes [dir];
  if (d)
    do_unlink_got_dir (x, phase, d);
  else {
    sobj->lookup_fmap
      (dir,
       wrap (mkref (this),
	     &bob::do_unlink_got_dir, x, phase));
  }
}

void
bob::do_unlink_got_dir (xrp x, int phase, snode* d)
{
  if (!d || d->ent->inode->type != CFSDIR) {
    warn << "unlink: cannot find directory\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
  cfs_inode_reg *dnode = d->ent->inode->reg;
  if (dnode->direct.size () == 0) {
    done_unlink (x, phase, true);
    return;
  } 
  
  bigint dirblk = dnode->direct[0];
  snode *dbe = sobj->active_snodes [dirblk];
  if (dbe)
    do_unlink_got_dirblk (x, phase, 0);
  else
    sobj->lookup_fmap
      (dirblk,
       wrap (mkref (this), &bob::do_unlink_fetch_dirblk_cb,
	     wrap (mkref (this), &bob::do_unlink_got_dirblk,
	           x, phase, 0)));
}

void
bob::do_unlink_fetch_dirblk_cb
  (callback<void,void>::ref cb, snode *n)
{
  if (!n || n->ent->type != CFS_DIRBLK) {
    warn << "unlink: directory block does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
  cb ();
}

void
bob::do_unlink_got_dirblk (xrp x, int phase, int blknum)
{
  bigint dir = unlink_parse_dir (x->record, phase);
  snode *d = sobj->active_snodes [dir];
  assert (d);
  cfs_inode_reg *dnode = d->ent->inode->reg;
  bigint dirblk = dnode->direct[blknum];
  snode *dbe = sobj->active_snodes [dirblk];
  assert (dbe);
  remove_dirent (x, phase, d, dbe->ent, blknum);
}

void 
bob::remove_dirent (xrp x, int phase, snode *d,
                                     cfs_data *dir, uint32 blknum)
{
  cfs_inode_reg *dnode = d->ent->inode->reg;
  assert (dirop_record (x->record));
  str name = unlink_parse_name (x->record, phase);

  bigint fh;
  cfs_dirent* de = dir->dir->entries;
  if (de && (de->name == name)) {
    fh = de->fh;
    dir->dir->entries = dir->dir->entries->nextentry;
  }
  else {
    while (de && de->nextentry && (de->nextentry->name != name))
      de = de->nextentry;
    if (!de || !de->nextentry) {
      if (blknum == dnode->direct.size ()-1)
        done_unlink (x, phase, true);
      else {
        blknum++;
	bigint dirblk = dnode->direct[blknum];
        snode *dbe = sobj->active_snodes [dirblk];
        if (dbe)
          do_unlink_got_dirblk (x, phase, blknum);
        else
          sobj->lookup_fmap
            (dirblk, 
	     wrap (mkref (this),
	           &bob::do_unlink_fetch_dirblk_cb,
	           wrap (mkref (this),
		         &bob::do_unlink_got_dirblk,
		         x, phase, blknum)));
      }
      return;
    }
    fh = de->nextentry->fh;
    de->nextentry = de->nextentry->nextentry;
  }

  assert(dir->dir->count > 0);
  dir->dir->count--;

  bigint obj = dirop_record (x->record)->dir ();
  if (incr_revision (&dnode->attr, &dnode->history, x)) {
    treeop_fail (CFSRWERR_VERCONFL);
    outstanding_treeops--;
    next ();
    return;
  }

  snode *ie = sobj->active_snodes [fh];
  if (ie)
    remove_dirent_got_fh (x, phase, fh, dnode, ie);
  else {
    sobj->lookup_fmap
      (fh, wrap (mkref (this), &bob::remove_dirent_got_fh,
		 x, phase, fh, dnode));
  }
}

void
bob::remove_dirent_got_fh (xrp x, int phase, bigint fh,
                                            cfs_inode_reg *dnode, snode *node)
{
  if (!node || node->ent->type != CFS_INODE) {
    warn << "unlink: object does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  cfs_inode *inode = (node->ent)->inode;
  if (inode->type == CFSLNK)
    inode->lnk->attr.nlink--;
  else {
    inode->reg->attr.nlink--;
  
    rename_delta *p = x->record->TEMPLATE record<rename_delta>();
    // dont need to check if dir is empty if we are renaming dir
    if (inode->type == CFSDIR && p == 0L)  {
      // check if directory is empty
      unlink_check_dir_empty 
	(inode,
	 wrap (mkref (this),
	       &bob::unlink_dir_empty, x, phase, dnode));
      return;
    }
  }
  done_unlink (x, phase);
}

void
bob::unlink_check_dir_block 
  (cfs_inode *inode, cbb cb, uint32 blocknum, ptr<cfs_data> data)
{
  if (!data || data->type != CFS_DIRBLK) {
    warn << "unlink: cannot retrieve data, directory may not be empty\n";
    cb (false);
    return;
  }
  
  cfs_dirent *e = data->dir->entries;
  if (e) {
    cb (false);
    return;
  }

  if (inode->reg->direct.size() > blocknum+1) {
    blocknum++;
    bigint dirblock = inode->reg->direct[blocknum];
    ns_retrieve_cfs_readonly
      (dirblock,
       wrap (mkref (this), &bob::unlink_check_dir_block,
	     inode, cb, blocknum));
  }
  else
    cb (true);
}

void
bob::unlink_check_dir_empty (cfs_inode *inode, cbb cb)
{
  if (inode->reg->direct.size() > 0) {
    bigint dirblock = inode->reg->direct[0];
    ns_retrieve_cfs_readonly
      (dirblock,
       wrap (mkref (this), 
	     &bob::unlink_check_dir_block, inode, cb, 0));
  }
  else
    cb (true);
}

void
bob::unlink_dir_empty (xrp x, int phase, cfs_inode_reg *dnode,
                                        bool empty)
{
  if (!empty) {
    warn << "cannot remove directory: not empty\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
  // remove ..'s link count in parent dir
  dnode->attr.nlink--;
  done_unlink (x, phase);
}

void
bob::done_unlink (xrp x, int phase, bool noentry)
{
  outstanding_treeops--;
  rename_delta *r = x->record->TEMPLATE record<rename_delta>();  
  if (r) {
    if (phase == 0)
      do_unlink (x, 1);
    else
      do_link (x);
  }
  else {
    if (noentry) {
      warn << "unlink failed: cannot find entry to remove\n";
      treeop_fail (CFSRWERR_ERROR);
    }
    next ();
  }
}

void 
bob::setattr (xrp x)
{
  outstanding_treeops++;
  attr_delta *a = x->record->TEMPLATE record<attr_delta> ();
  snode *e = sobj->active_snodes [a->object ()];
  if (e)
    setattr_got_node (x, e);
  else {
    sobj->lookup_fmap
      (a->object (),
       wrap (mkref (this), &bob::setattr_got_node, x));
  }
}

void
bob::setattr_got_node (xrp x, snode *e)
{
  attr_delta *a = x->record->TEMPLATE record<attr_delta> ();
  if (!e || e->ent->type != CFS_INODE) {
    warn << "cannot set attribute: object does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }
    
  if (e->ent->inode->type == CFSLNK) {
    cfs_inode_lnk *lnode = e->ent->inode->lnk;
    set_attr (lnode, a->attr ());
    
    if (a->attr()->valid_mode) {
      if (incr_revision (&lnode->attr, &lnode->history, x, true)) {
	treeop_fail (CFSRWERR_VERCONFL);
        outstanding_treeops--;
        next ();
        return;
      }
    }

    outstanding_treeops--;
    next ();
  }
  else {
    cfs_inode_reg *rnode = e->ent->inode->reg;

    if (a->attr()->valid_mode) {
      if (incr_revision (&rnode->attr, &rnode->history, x, true)) {
	treeop_fail (CFSRWERR_VERCONFL);
        outstanding_treeops--;
        next ();
        return;
      }
    }

    set_attr (rnode, a->attr ());

    outstanding_treeops--;

    // write_file calls incr_revision if size of the file is changed
    write_file (x, rnode);
    if (outstanding_treeops == 0)
      next ();
  }
}

void
bob::do_rename (xrp x)
{
  do_unlink (x, 0);
}


extern void
cfsrw_read_file (int fd, bigint fh, uint64 size,
                 ptr<snapshot_obj> sobj, cbcfsrw cb);

void 
bob::do_patch (xrp x)
{
  outstanding_treeops++;
  
  patch_delta *p = x->record->TEMPLATE record<patch_delta>();
  snode *f = sobj->active_snodes [p->file ()];
  if (f)
    do_patch_got_node (x, f);
  else
    sobj->lookup_fmap
      (p->file (), wrap (mkref (this), &bob::do_patch_got_node, x));
}

void
bob::do_patch_got_node (xrp x, snode *f)
{
  if (!f || f->ent->type != CFS_INODE) {
    warn << "patch: file does not exist\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  cfs_inode *inode = (f->ent)->inode;
  if (inode->type != CFSREG) {
    warn << "patch: file not a regular file\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  if (inode->reg->attr.nlink == 0) {
    warn << "cannot patch file with link count 0\n";
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  patch_delta *p = x->record->TEMPLATE record<patch_delta>();
  uint64 nsz = inode->reg->attr.size;

  str s = pastwatch_tmpfile ();
  int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);

  cfsrw_read_file (fd, p->file (), nsz, sobj,
                   wrap (mkref (this), &bob::do_patch_read_cb, s, fd, x));
}

extern "C" {
  int patch_run (int argc, char **argv);
};

static int
patch (const char *f, const char *p)
{
  int argc = 6;
  const char *argv [6] = { "patch", "-s", "-f", "-i", p, f };
  int r = patch_run (argc, (char**) argv);
  return r;
}

extern void
cfsrw_write_file_local
(str path, bigint fh, bigint va,
 callback<void, cfsw_write_arg, cfsrw_status>::ref cb);

void
bob::do_patch_read_cb (str tfn, int rfd, xrp x, cfsrw_status status)
{
  patch_delta *p = x->record->TEMPLATE record<patch_delta>();
  close (rfd);

  if (status) {
    warn << "patch: can't read file " << armor64A (p->file ()) << "\n";
    unlink (tfn);
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  str px = p->patch ();
  str pf = pastwatch_tmpfile ();
  int fd = open (pf, O_CREAT | O_TRUNC | O_WRONLY, 0600);
  unsigned n = 0;

  while (n < px.len ()) {
    int r = write (fd, px.cstr ()+n, px.len ());
    if (r <=0)
      break;
    n += r;
  }
  close (fd);

  if (n < px.len ()) {
    warn << "patch: can't write patch to file\n";
    unlink (tfn);
    unlink (pf);
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  int r = patch (tfn, pf);
  unlink (pf);
  
  if (r != 0) {
    warn << "patch: can't patch file " << armor64A (p->file ()) << "\n";
    warn << "orig and rej files have prefix " << tfn << "\n";
    // unlink (tfn);
    treeop_fail (CFSRWERR_ERROR);
    outstanding_treeops--;
    next ();
    return;
  }

  cfsrw_write_file_local
    (tfn, p->file (), viewaddr,
     wrap (mkref (this), &bob::do_patch_write_patch, tfn, x));
}

void
bob::do_patch_write_patch (str tfn, xrp x, cfsw_write_arg a,
                           cfsrw_status status)
{
  outstanding_treeops--;
  unlink (tfn);

  if (status) {
    warn << "patch: can't write patched file\n";
    treeop_fail (CFSRWERR_ERROR);
    next ();
    return;
  }

  patch_delta *p = x->record->TEMPLATE record<patch_delta>();
  snode *f = sobj->active_snodes [p->file ()];
  assert (f);

  cfs_inode_reg *rnode = f->ent->inode->reg;
  rnode->direct.clear ();
  rnode->direct.setsize (a.direct.size ());
  for (unsigned i=0; i < a.direct.size (); i++)
    rnode->direct[i] = a.direct [i];
  rnode->indirect = a.indirect;
  rnode->double_indirect = a.double_indirect;
  rnode->triple_indirect = a.triple_indirect;
  rnode->attr.size = a.size;

  if (incr_revision (&rnode->attr, &rnode->history, x))
    treeop_fail (CFSRWERR_ERROR);
  next ();
}

