/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "rxx.h"
#include "ns.h"
#include "record.h"
#include "cfsw_record.h"
#include "cfsw_prot.h"

ptr<ns> nsobj = 0;
int view_debug = 0;
int ns_debug = 0;
int cfs_debug = 0;

extern "C" {
time_t get_date (char *raw);
};

class mklog {
private:
  static int work;

  static bigint
  make_block_id (const void *data, int size) 
  {
    long *rd = (long *)data;
    char id[sha1::hashsize];
    sha1_hash (id, rd, size);
    
    bigint ID;
    mpz_set_rawmag_be (&ID, id, sizeof (id));
    return ID;
  }

  static void
  store_block_cb (callback<void, bigint, int>::ref cb,
                  ns_stat status, bigint key)
  {
    if (status != NS_OK) {
      warn << "store_block_cb failed\n";
      exit(1);
    }
    cb (key, 1);
  }

  static void
  store_cb (bool ispk, bigint id, int ok)
  {
    if (ispk) {
      char raw[20];
      mpz_get_rawmag_be (raw, sizeof (raw), &id);
      str name = armor64A (raw, sizeof (raw));
      strbuf out;
      out << name << "\n";
      out.tosuio ()->output (1);
    }
  
    work--;
    if (work == 0)
      exit (0);
  }

  static bigint
  store (bigint va, const str &data, callback<void, bigint, int>::ref cb)
  {
    return nsobj->insert_block
      (data.cstr(), data.len(), wrap(store_block_cb, cb));
  }

  static void
  store_lptr (bigint va, const str &data,
              callback<void, bigint, int>::ref cb, ptr <sfspriv> sk)
  {
    assert (sk);
    sfs_sig2 s;
    sk->sign (&s, data);
    sfs_pubkey2 k;
    sk->export_pubkey (&k);
    nsobj->insert_pk (sk, data, wrap (store_block_cb, cb));
  }

public:
  
  static void
  create_log (ptr <sfspriv> sk, bigint va, bool root, str timestamp)
  {
    bigint cid = 0;
    bigint head = ns::pkid (sk);
 
    if (root) {
      cfsw_sattr attr;
      attr.valid_mode = 1;
      attr.mode = 0755;
      attr.valid_size = 1;
      attr.size = 0;
      inode_delta r (CFSW_DIR, attr);

      work++;
      bigint rid = store (va, r.serialize (), wrap (store_cb, false));

      commit_record c;
      //      c._proposed_name = 
      c._head = head;
      c._prev = 0;
      c._ndeltas = 1;
      c._deltas.push_back (rid);
      if (timestamp && timestamp != "") {
	time_t t = get_date ((char*) timestamp.cstr());
	c._time = t;
      }

      work++;
      cid = store (va, c.serialize (), wrap (store_cb, false));
    }

    log_pointer p;
    p._seqn = 2;
    p._commit = cid;
    p._lock = 0;
    p._verf = 0;

    work++;
    store_lptr (va, p.serialize (), wrap (store_cb, true), sk);
  }
};

int mklog::work = 0;

static void
usage ()
{
  warnx << "usage: " << progname << " [-S timestamp] [-s <keyfile>] [-c] proj\n";
  exit (1);
}

void
start (bool create, bigint va, ptr<sfspriv> sk, str timestamp)
{
  mklog::create_log (sk, va, create, timestamp);
}

int
main (int argc, char **argv)
{
  char *sk_file = 0;
  bool create = false;
  str timestamp = "";

  setprogname (argv[0]);
  random_init ();

  argv++;
  argc--;

  if (argc < 1) 
    usage();
  
  while ( !strcmp(argv[0], "-s") || !strcmp(argv[0], "-c") || !strcmp(argv[0], "-S")) {

    if (!strcmp(argv[0], "-s")) {
      sk_file = argv[1];
      argv += 2;
      argc -= 2;
      if (argc == 0) 
	break;
    }
    
    if (!strcmp(argv[0], "-c")) {
      create = true;
      argv ++;
      argc --;
      if (argc == 0)
	break;
    }

    if (!strcmp(argv[0], "-S")) {
      timestamp = argv[1];
      argv += 2;
      argc -= 2;
      if (argc == 0)
	break;
    }
  }

  if ((argc != 1) || !sk_file)
    usage ();

  bigint va;
  char sym[1024];
  str dearm;
  rxx rx ("project:([0-9a-zA-Z+\\-]*)[/]*$");
  int r = readlink(argv[0], sym, 1024);
  if (r > 0) sym[r] = '\0';
  if (rx.search (argv[0]) || (r > 0 && rx.search (sym))) {
    dearm = dearmor64A(rx [1]);
  } else {
    dearm = dearmor64A(argv [0]);
  }
  if (dearm) 
    mpz_set_rawmag_be(&va, dearm.cstr(), dearm.len());
  else {
    warnx << "pastmklog: invalid project ID\n";
    exit(1);
  }

  ptr <sfspriv> sk;
  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    exit (1);
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    exit (1);
  }

  nsobj = ns::alloc (va);
  nsobj->connect (wrap (start, create, va, sk, timestamp));
  amain();
}

