
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "replicaimpl.h"
#include "penviron.h"

void
replica_meta::serialize (strbuf &sb) const
{
  pkv.serialize (sb);
  commits.serialize (sb);
  suint (sb, addrs.size ());
  for (unsigned i=0; i<addrs.size (); i++)
    sbigint (sb, addrs [i]);
}

bool
replica_meta::parse (xin &x)
{
  if (!vervec::parse (&x, pkv))
    return false;
  if (!comvec::parse (&x, commits))
    return false;
  unsigned n;
  if (!x.ruint (n))
    return false;
  for (unsigned i=0; i<n; i++) {
    bigint a;
    if (!x.rbigint (a))
      return false;
    addrs.push_back (a);
  }
  return true;
}

replica_impl::replica_impl (bigint v, bigint ra)
  : _view (v), _replicaaddr (ra)
{
  _mybranch = -1;
  _changed = false;
  
  _rdb = open_rdb ();
  _btree = 0;
  if (_rdb)
    _btree = New refcounted<btree_srv> (_rdb);
}

void
replica_impl::dispatch_with_snapshot (cfsrcallp c, sptr s,
                                      cfsrw_status status)
{
  if (status) {
    c->reply (CFSRWERR_NOREPLICA);
    return;
  }

  c->_s = s;
  cfsr::dispatch (c);
}

void
replica_impl::dispatch (cfsrcallp c)
{
  if (c->saddr == 0) {
    assert (_mybranch >= 0);
    c->_s = _branches [_mybranch];
    cfsr::dispatch (c);
  }
  else
    find (c->saddr, wrap (this, &replica_impl::dispatch_with_snapshot, c));
}

void 
replica_impl::sync (cbaddr cb)
{
  if (_branches.size ()) {
    assert (_rdb);
  
    strbuf b;
    _m.serialize (b);
    str s = b;
  
    ptr<dbrec> k = bigint2dbrec (_replicaaddr);
    ptr<dbrec> v = str2dbrec (s);
    _rdb->remove (k);
    _rdb->insert (k, v);
    _rdb->sync ();
 
    _changed = false;

    cb (_replicaaddr, CFSRW_OK);
  }
  else
    cb (0, CFSRWERR_NOREPLICA);
}

str
replica_impl::branch_tag (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->btag ();
}

bigint
replica_impl::branch_head_snapshot (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->addr ();
}

commit_info
replica_impl::branch_head_info (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->branch ();
}

int
replica_impl::my_branch () const
{
  return _mybranch;
}

bool
replica_impl::set_branch_head (bigint n)
{
  unsigned index = get_branch_index(n);

  if (index < 0) 
    return false;
  else {
    _mybranch = index;
    return true;
  }
}

int 
replica_impl::get_branch_index (bigint n) const
{
  assert (_branches.size () > 0);
  for (unsigned i=0; i<_branches.size (); i++) {
    if (branch_head_info (i).rptr == n) {
      return i;
    }
  }
  return -1;
}

bigint
replica_impl::root () const
{
  assert (_branches.size () > 0);
  return _branches [0]->root ();
}

static inline void
sort_snapshots (spv& v)
{ 
  if (v.empty ())
    return;
  spv oldv = v;
  v.clear ();
  
  while (v.size () < oldv.size ()) {
    int x = -1;
    bigint h;
    for (unsigned i=0; i<oldv.size (); i++) {
      if (oldv [i] == 0)
        continue; 
      if (x == -1 || oldv [i]->addr () > h) {
        h = oldv [i]->addr ();
        x = i;
      }
    }
    assert (x >= 0);
    v.push_back (oldv [x]);
    oldv [x] = 0;
  }
}

void
replica_impl::set_addrs (spv nb)
{
  sort_snapshots (nb);
  int myb = -1;
  bool forked = false;

  _branches.clear ();
  for (unsigned i=0; i<nb.size (); i++) {
    ptr<snapshot> s = obj2impl (nb [i]);
    if (s->flags () & commit_record::IGNORE_BRANCH)
      continue;
    if (_mybmagic != "" && s->ptag == _mybmagic) {
      if (myb == -1 && !forked)
        myb = i;
      else if (myb != -1) {
	warn << "user specified branch forked\n";
	forked = true;
	myb = -1;
      }
    }
    _branches.push_back (s);
  }

  vec<bigint> v;
  for (unsigned i=0; i<_branches.size (); i++) {
    ptr<snapshot> si = obj2impl (_branches [i]);
    v.push_back (_branches [i]->addr ());
  }
  _m.addrs = v; // _m is a list of snapshots of the branch leaves.

  if (myb >= 0)
    _mybranch = myb;
  else
    _mybranch = -1;
}

static inline void
add_snapshot_to_list (sptr s, spv& v)
{
  bigint na = s->addr ();
  bool found = false;
  for (unsigned i=0; i<v.size (); i++) {
    if (v [i]->addr () == na) {
      found = true;
      v [i] = s;
    }
  }
  if (!found)
    v.push_back (s);
}

// loads a set of snapshots as the latest snapshots for the
// repository. at the minimum a base snapshot is loaded (base = true).
// if there are branches, they are loaded as well.

void
replica_impl::load_next (vec<bigint> v, cbcfsrw cb,
                         sptr s, cfsrw_status status)
{
  if (!status) {
    assert (s);
    ptr<snapshot> si = obj2impl (s);
    _branches.push_back (s);
    add_snapshot_to_list (s, _slist);

    if (!v.empty ()) {
      bigint a = v.pop_front ();
      load_snapshot (a, wrap (this, &replica_impl::load_next, v, cb));
    }
    else {
      // loaded all snapshots
      cb (CFSRW_OK);
    }
  }
  else {
    _branches.clear ();
    cb (status);
  }
}

void
replica_impl::load (cbcfsrw cb)
{
  if (!_rdb) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  ptr<dbrec> k = bigint2dbrec (_replicaaddr);
  ptr<dbrec> d = _rdb->lookup (k);
  if (!d) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  str mx = dbrec2str (d);
  xin x (mx);
  if (!_m.parse (x)) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  if (_m.addrs.size () == 0) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  _branches.clear ();
  vec<bigint> v = _m.addrs;
  bigint a = v.pop_front ();
  load_snapshot (a, wrap (this, &replica_impl::load_next, v, cb));
}

ptr<pastdbfe>
replica_impl::open_rdb ()
{
  str r = pget_replicadir();
  str d = r << "/" << armor64A (_replicaaddr);
  mkdir (r, 0755);

  dbOptions opts;
  opts.addOption("opt_flag", DB_CREATE);
  opts.addOption ("opt_cachesize", 1000);
  opts.addOption ("opt_nodesize", 4096);
  ptr<pastdbfe> db = New refcounted<pastdbfe> ();
  if (db->opendb (d.cstr (), opts))
    return 0;
  return db;
}

bool
replica_impl::lookup_snapshot (bigint c, bigint &s)
{
  if (!_btree){
    return false;
  }
  btree_node n;
  if (!_btree->lookup_by_commit (c, n)){
    return false;
  }
  s = n.commit.vptr;
  return true;
}

// find a particular snapshot using its address. could return a cached
// snapshot object, or load the snapshot if it does not exists.

void
replica_impl::load_cb (cbloadsn_t cb, sptr s, cfsrw_status status)
{
  if (!status) {
    assert (s);
    add_snapshot_to_list (s, _slist);
  }
  cb (s, status);
}

static void
find_cb (cbloadsn_t cb, sptr s)
{
  cb (s, CFSRW_OK);
}

void
replica_impl::find (bigint addr, cbloadsn_t cb)
{
  for (unsigned i=0; i<_slist.size (); i++) {
    if (_slist [i]->addr () == addr) {
      delaycb (0, wrap (find_cb, cb, _slist[i]));
      return;
    }
  }
  load_snapshot
    (addr, wrap (this, &replica_impl::load_cb, cb));
}

// builds snapshots, either from scratch, if no base snapshot is
// loaded, or from previously built snapshots.

void
replica_impl::build_cb (comvec commits, cbcfsrw cb, spv v, cfsrw_status status)
{
  // yipal: I think <v> should contain an entry for each new snapshot.

  if (status && status != CFSRWERR_NONEWRECS) {
    warn << "cannot build snapshot, err " << status << "\n";
    _branches.clear ();
    _mybranch = -1;
    cb (status);
    return;
  }

  _changed = true;
  nsobj->sync ();

  if (status == CFSRWERR_NONEWRECS) {
    // warn << "no new deltas, no new snapshots built\n";
    _m.commits = commits;
    v.clear ();
    v = _branches;
  }
  else {
    assert (v.size () > 0);
    for (unsigned j=0; j<v.size (); j++)
      add_snapshot_to_list (v[j], _slist);
    _m.commits = commits;
  }

  set_addrs (v);
  cb (CFSRW_OK);
}

void
replica_impl::update (comvec commits, cbcfsrw cb)
{
  if (!_rdb) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  if (_mybranch >= 0) {
    char xxb[20];
    rnd.getbytes (xxb, 20);
    _mybmagic = str (xxb, 20);
    ptr<snapshot_obj> s = _branches [_mybranch];
    _mybptag = obj2impl (s)->ptag;
    obj2impl (s)->ptag = _mybmagic;
  }
  else
    _mybmagic = "";

  if (_branches.size () > 0)
    build_snapshot
      (this, _view, commits,
       wrap (this, &replica_impl::build_cb, commits, cb), _branches);
  else
    build_snapshot
      (this, _view, commits,
       wrap (this, &replica_impl::build_cb, commits, cb));
}

void
replica_impl::load_old_cb (cbcfsrw cb, sptr s, cfsrw_status status)
{
  if (status == CFSRW_OK) {
    _branches.push_back (s);
    add_snapshot_to_list (s, _slist);
    cb (CFSRW_OK);
  }
  else
    cb (status);
}

void
replica_impl::load_old_commit (bigint c, cbcfsrw cb)
{
  bigint s;
  if (!lookup_snapshot (c, s))
    cb (CFSRWERR_STALE);
  else
    load_snapshot
      (s, wrap (this, &replica_impl::load_old_cb, cb));
}

bigint
replica_impl::determine_current_branch(str arg_branch, bool ignore_implicit_branches, 
				       str crec_from_info) const
{
  // Find and set current branch
  // -- Checks the various forms that arg_branch can take.
  // if we can't find a branch with arg_branch, then try the
  // commit_record from the info file if the user is ignoring implicit
  // forks.
  bigint c=0;
  str current_guess = arg_branch;

  // Compare current_guess to current branch heads.
  if (current_guess && current_guess != "") {
    str n;
    vec<bigint> matching_branch_heads;

    for (unsigned i=0; i<nbranches (); i++) {
      if (current_guess == branch_tag (i)) {
	c = branch_head_info (i).rptr;
	matching_branch_heads.push_back(c);
      }
    }

    if (matching_branch_heads.size() == 1) {
      return matching_branch_heads[0];

    } else if (matching_branch_heads.size() > 1 && ignore_implicit_branches && 
	       crec_from_info && crec_from_info != "") {
      // read commit_record addr from info file and search for a branch head
      // that is a sole descendant of the commit_record.
      bigint cr;
      bigint head;
      //str commit_rec = get_commit_record_from_infofile(file);
      if (dearmor64A_bigint (crec_from_info, cr)) {
	if (_btree->get_sole_descendant (cr, head)) {

	  // Check that the branchtags match.
	  str headtag = branch_tag (get_branch_index(head));
	  if (headtag != current_guess) {
	    warn << "Chose head " << armor64A(head) << "(" << headtag 
		 << ") because of commit record " << armor64A(cr) 
		 << " in Pastwatch/info, but branch tag \"" 
		 << current_guess << "\" does not match.\n";
	    return 0;
	  }
	  return head;
	}
      }
    } else if (matching_branch_heads.size() > 1) {
      warn << "Branch \"" << current_guess << "\" has forked.  You must specify a "
	   << "branch head to follow with the -b option.\n";
      return 0;
    }
  }
  
  // Compare current_guess to tagged versions
  if (current_guess && current_guess != "") {
    if (_btree->lookup_by_tag(current_guess, c)){
      return c;
    }
  }

  // Check if current_guess is equal to the branch's commit_record key.
  if (current_guess && current_guess != "" && 
      dearmor64A_bigint (current_guess, c)){
    return c;
  }

  // Check if current_guess is undefined, but there is only one branch
  if (!current_guess || current_guess == "") {
    if (nbranches () == 1) {
      return branch_head_info (0).rptr;
    }
  }

  return c;
}
