/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@mit.edu),
 *                    Massachusetts Institute of Technology
 * 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#ifndef _DHASH_BLOCKSTORE_H_
#define _DHASH_BLOCKSTORE_H_

// probably have to support db3.h and adb btree like dbfe does.
#include <db.h> 
#include "dbfe.h"
#include "bigint.h"

class bsEnumeration {
  DBC *cursor;

public:
  bsEnumeration(DB *db);
  ~bsEnumeration();
  
  void close(); // free the cursor
  ptr<dbrec> nextElement();
  ptr<dbrec> nextElement(ptr<dbrec> startkey);
};

class dhash_blockstore {
 public:
  dhash_blockstore();
  ~dhash_blockstore();

  int opendb(str dirname, bool createenv, time_t synctime);
  int insert_pkblock(bigint projkey, ptr<dbrec> key, ptr<dbrec> data);
  int insert_chblock(bigint projkey, ptr<dbrec> key, ptr<dbrec> data);
  ptr<dbrec> lookup_pkblock(ptr<dbrec> key, bigint *projkey = NULL);
  ptr<dbrec> lookup_chblock(ptr<dbrec> key, bigint *projkey = NULL);
  int ch_get_projkey(ptr<dbrec> key, bigint *projkey);
  int remove_pkblock(ptr<dbrec> key);
  int remove_chblock(ptr<dbrec> key);

  // Dont use a stateful enumerator, because it would lock the db tables.
  int getpkblockids(vec<ptr<dbrec> > *rec_list, bigint *projkey = NULL);
  int getchblockids(vec<ptr<dbrec> > *rec_list, bigint *projkey = NULL);
  int getchblockids(vec<ptr<dbrec> > *rec_list, ptr<dbrec> start, ptr<dbrec> end, uint32_t max);

  ptr<bsEnumeration> enumerate_ch() { return New refcounted<bsEnumeration>(block_db.metadata); }

  void sync_cb();

  void dump_pk(str message, bigint *projkey = NULL);
  void dump_ch(str message, bigint *projkey = NULL);

 private:
  struct db_agg {
    // Aggregates 3 databases. One for data, one for metadata, and
    // one mapping project id to block ids.

    DB *data;
    DB *proj2blocks;
    DB *metadata; // maps a block id to a metadata_entry

    void init() {  data = proj2blocks = metadata = NULL; }
    void close() {
      if (data) 
	data->close(data, 0);
      if (proj2blocks) 
	proj2blocks->close(proj2blocks, 0);
      if (metadata) 
	metadata->close(metadata, 0);
    }
    void sync() {
      data->sync(data, 0L);
      proj2blocks->sync(data, 0L);
      metadata->sync(data, 0L);
    }
  };

  struct metadata_entry {
    char rawprojkey[20];
    void setprojkey(bigint key) {
      bzero(&rawprojkey, 20);
      mpz_get_rawmag_be((char*) &rawprojkey, 20, &key);
    }
    bigint getprojkey() {
      bigint projkey;
      str s((char*)&rawprojkey, 20);
      mpz_set_rawmag_be( &projkey, s.cstr(), s.len() );
      return projkey;
    }
  };

  DB_ENV *dbe;
  db_agg block_db, keyhash_db;
  time_t synctime_;

  int open_db(DB **db, DB_ENV *dbe, char *filename, int preflags, int flags);
  int insert_block(DB *db, DB_TXN *txn, bigint *projkey, ptr<dbrec> key, 
		   ptr<dbrec> data);
  int insert_trackedblock(db_agg *agg, bigint projkey, ptr<dbrec> key, ptr<dbrec> data);
  int getblockids_from(db_agg *agg, vec < ptr<dbrec> > *rec_list, bigint *projkey);
  ptr<dbrec> lookup_block(DB *db, ptr<dbrec> key, bigint *projkey);
  int remove_trackedblock(db_agg *agg, ptr<dbrec> key);


  // dst->data must be cleaned up by caller for both dbrec2dbt and bigint2dbt
  int dbrec2dbt(ptr<dbrec> src, DBT *dst, bigint *projkey = NULL);
  int bigint2dbt(bigint src, DBT *dst);

  ptr<dbrec> bigint2dbrec(bigint src);
  ptr<dbrec> dbt2dbrec(DBT *src, bigint *projkey = NULL);
 };

#endif

