/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "ns.h"
#include "pack.h"
#include "rxx.h"
#include <time.h>
#include "view.h"
#include "record.h"
#include "logw.h"

typedef callback<void, ptr<view_state> >::ptr cbex;

class logw_impl : public virtual logw
{
private:
  bigint _head;
  ref<sfspriv> _sk;
  ptr<view> _view;
  bigint _viewaddr;
  bigint _verf;
  comvec _commits;
  vervec _pkv;

  lptr _lptr;
  bigint _pcommit;
  unsigned _pseqn;
  vec<bigint> _deltas;
  bool _commit;
  bigint _commit_addr;
  unsigned _work;
  unsigned _cerr;

  callback<void,void>::ptr _synccb;
  int _stores;
  int _failed_stores;

  void reset ();
  void store_cb (ns_stat, bigint);
  void store_lptr (cbl cb, ns_stat, bigint);
  void abort_cb (cbl cb, log_status);
  void sync_lptr (cbl cb);

  void excl_fetch (cbex xcb, cbx cb, ptr<view_state> v, log_status s);
  void excl_rlock (bigint h, cbx cb, ptr<view_state> v);
  void excl_first (bigint h, cbx cb, ptr<view_state> v);
  void excl_write (bigint h, cbx cb, log_status status);
  void excl_second (bigint h, cbx cb, ptr<view_state> v);
  void excl_zombie (bigint h, cbx cb, log_status status);

  void commit_abort (cbc cb, log_status status);
  void commit_cb (cbc cb, ns_stat status, bigint n);
  void commit_sync_cb (cbc cb, log_status status);
  bigint make_sind (ptr<commit_record> c, cbc cb);

public:
  logw_impl (bigint la, ptr<sfspriv> sk, bigint verf, ptr<view> v, bigint va);
  bigint head () const { return _head; }
  unsigned seqn () const { assert (_lptr); return _lptr->seqn (); }
  void append (deltap);

  void rlock   (bigint h, const comvec& v, const vervec& pkv, cbx cb);
  void prepare (bigint h, const comvec& v, const vervec& pkv, cbx cb);
  void rlock   (bigint h, cbx cb);
  void prepare (bigint h, cbx cb);
  void commit  (ptr<commit_record> c, cbc cb);
  void abort   (cbl cb);
};

ptr<logw>
logw::alloc (bigint la, ptr<sfspriv> sk, bigint verf,
             ptr<view> v, bigint va)
{
  return New refcounted<logw_impl> (la, sk, verf, v, va);
}

logw_impl::logw_impl (bigint la, ptr<sfspriv> sk, bigint verf,
                      ptr<view> v, bigint va)
  : _head (la), _sk (sk), _view (v), _viewaddr (va), _verf (verf)
{
  reset ();
  _lptr = 0;
  _pcommit = 0;
}

void
logw_impl::reset ()
{
  _stores = 0;
  _failed_stores = 0;
  _synccb = 0;
  _commit = false;
  _deltas.clear ();
  _work = 0;
  _cerr = 0;
}

void
logw_impl::append (deltap r)
{
  _stores++;
  str data = r->serialize();
  bigint n = nsobj->insert_block
    (data.cstr (), data.len (), wrap (mkref (this), &logw_impl::store_cb));
  _deltas.push_back (n);
}

void
logw_impl::store_cb (ns_stat status, bigint n)
{
  _stores--;
  if (status)
    _failed_stores++;
  if (_stores == 0 && _synccb) {
    cbv cb = _synccb;
    _synccb = 0;
    cb ();
  }
}

void
logw_impl::store_lptr (cbl cb, ns_stat status, bigint n)
{
  if (status)
    cb (LOGERR_IO);
  else
    cb (LOG_OK);
}

void
logw_impl::sync_lptr (cbl cb)
{
  assert (_lptr);
  if (_stores == 0) {
    if (_failed_stores) {
      abort (cb);
      return;
    }
    str data = _lptr->serialize();
    nsobj->insert_pk
      (_sk, data, wrap (mkref (this), &logw_impl::store_lptr, cb));
  }
  else
    _synccb = wrap (mkref (this), &logw_impl::sync_lptr, cb);
}

void
logw_impl::excl_second (bigint h, cbx cb, ptr<view_state> v)
{
  assert (v);
  
  for (unsigned i=0; i<v->lptrs.size (); i++) {
    if (!v->lptrs [i])
      // ns.C didn't load a commit pointer, probably because the
      // public key block has not changed since the last time we saw
      // it. this is okay, since we must not have seen a lock here
      // last time because we would have aborted.
      continue;
    
    if (_view->head (i) == _head && v->lptrs [i]->_verf != 0 &&
	v->lptrs [i]->_verf != _verf) { // zombie lock
      warn << "found zombie again?\n";
      assert (0);
    }

    if (v->lptrs [i]->_lock == h && _view->head (i) != _head) {
      warn << armor64A (_view->head (i)) << " has lock\n";
      warn << "his verf is " << armor64A (v->lptrs [i]->_verf) << "\n";
      warn << "my verf is " << armor64A (_verf) << "\n";
      warn << "my user is " << armor64A (_head) << "\n";
      warn << "lock is " << armor64A (h) << "\n";
      cb (_commits, _pkv, LOGERR_EXCLCONT);
      return;
    }
  }

  // warn << "acquired mutual exclusion\n";
  cb (_commits, _pkv, LOG_OK);
}

void
logw_impl::excl_fetch (cbex xcb, cbx cb, ptr<view_state> v, log_status s)
{
  if (s) {
    warn << "excl_fetch error\n";
    cb (_commits, _pkv, LOGERR_IO);
  }

  else {
    assert (v);
    _commits = v->commits;
    _pkv.clear ();

    bool found = false;
    unsigned m = 0;
    unsigned i, j;
    for (i=0, j=0; i < v->lptrs.size ();) {
      if (j < v->pkv.size ()) {
	if (_view->head (i) == _head) {
	  m = i;
	  found = true;
	}
	if (_view->head (i) == v->pkv.head (j)) {
	  lptr c = v->lptrs [i];
          if (c && c->_verf == 0 && c->_lock == 0)
	    c = 0;
          if (!c)
	    // lptr did not change, or it has changed but it didn't
	    // specify a lock
	    _pkv.add (v->pkv.head (j), v->pkv.seqn (j));
	  i++;
	  j++;
	}
	else {
          assert (v->pkv.head (j) > _view->head (i));
	  i++;
	}
      }
      else
	break;
    }
    if (j != v->pkv.size ()) {
      warn << "v->pkv.size is " << v->pkv.size () << "\n";
      for (unsigned i=0; i<v->pkv.size (); i++) {
        warn << "h " << v->pkv.head (i) << ", " << v->pkv.seqn (i) << "\n";
      }
      assert (0);
    }

    assert (found);
    _lptr = v->lptrs [m];
    _pcommit = _lptr->commit ();
    _pseqn = _lptr->seqn ();
    assert (v->commits.head (m) == _head);
    xcb (v);
  }
}

void
logw_impl::excl_write (bigint h, cbx cb, log_status status)
{
  if (status)
    cb (_commits, _pkv, LOGERR_IO);
  else
    _view->fetch
      (_pkv, _commits,
       wrap (mkref (this), &logw_impl::excl_fetch,
	     wrap (mkref (this), &logw_impl::excl_second, h, cb), cb));
}

void
logw_impl::excl_zombie (bigint h, cbx cb, log_status status)
{
  if (status)
    cb (_commits, _pkv, status);
  else
    _view->fetch
      (_pkv, _commits,
       wrap (mkref (this), &logw_impl::excl_fetch,
             wrap (mkref (this), &logw_impl::excl_first, h, cb), cb));
}

void
logw_impl::excl_first (bigint h, cbx cb, ptr<view_state> v)
{
  assert (v);

  for (unsigned i=0; i<v->lptrs.size (); i++) {
    if (!v->lptrs [i])
      // ns.C didn't load a commit pointer, probably because the
      // public key block has not changed since the last time we saw
      // it. this is okay if we only ever pass to ns.C a vector that
      // captures versions of public key blocks w/o any locks.
      continue;
    
    if (_view->head (i) == _head && v->lptrs [i]->_verf != 0 &&
	v->lptrs [i]->_verf != _verf) { // zombie lock
      warn << "found zombie\n";
      _lptr->_verf = 0;
      _lptr->_lock = 0;
      sync_lptr (wrap (mkref (this), &logw_impl::excl_zombie, h, cb));
      return;
    }

    if (v->lptrs [i]->_lock == h) {
      warn << armor64A (_view->head (i)) << " has lock\n";
      warn << "his verf is " << armor64A (v->lptrs [i]->_verf) << "\n";
      warn << "my verf is " << armor64A (_verf) << "\n";
      warn << "my user is " << armor64A (_head) << "\n";
      warn << "lock is " << armor64A (h) << "\n";
      cb (_commits, _pkv, LOGERR_EXCLCONT);
      return;
    }
  }

  _lptr->_lock = h;
  _lptr->_verf = _verf;
  sync_lptr (wrap (mkref (this), &logw_impl::excl_write, h, cb));
}

void
logw_impl::excl_rlock (bigint h, cbx cb, ptr<view_state> v)
{
  cb (_commits, _pkv, LOG_OK);
}

void
logw_impl::prepare (bigint h, cbx cb)
{
  _view->fetch
    (wrap (mkref (this), &logw_impl::excl_fetch,
           wrap (mkref (this), &logw_impl::excl_first, h, cb), cb));
}

void
logw_impl::prepare (bigint h, const comvec& v, const vervec& pkv, cbx cb)
{
  _pkv.clear ();
  // always try to fetch our own log-head/commit pointer
  for (unsigned i=0; i<pkv.size (); i++)
    if (pkv.head (i) != _head)
      _pkv.add (pkv.head (i), pkv.seqn (i));
  _commits = v;

  _view->fetch
    (_pkv, _commits,
     wrap (mkref (this), &logw_impl::excl_fetch,
           wrap (mkref (this), &logw_impl::excl_first, h, cb), cb));
}

void
logw_impl::rlock (bigint h, cbx cb)
{
  _view->fetch
    (wrap (mkref (this), &logw_impl::excl_fetch,
           wrap (mkref (this), &logw_impl::excl_rlock, h, cb), cb));
}

void
logw_impl::rlock (bigint h, const comvec& v, const vervec& pkv, cbx cb)
{
  _pkv.clear ();
  // always try to fetch our own log-head/commit pointer
  for (unsigned i=0; i<pkv.size (); i++)
    if (pkv.head (i) != _head)
      _pkv.add (pkv.head (i), pkv.seqn (i));
  _commits = v;
  
  _view->fetch
    (_pkv, _commits,
     wrap (mkref (this), &logw_impl::excl_fetch,
           wrap (mkref (this), &logw_impl::excl_rlock, h, cb), cb));
}

void
logw_impl::commit_sync_cb (cbc cb, log_status status)
{
  if (status)
    cb (0, status);
  else
    cb (_lptr->_commit, LOG_OK);
  reset ();
}

void
logw_impl::commit_abort (cbc cb, log_status status)
{
  cb (0, status);
}

void
logw_impl::commit_cb (cbc cb, ns_stat status, bigint n)
{
  if (status)
    _cerr = 1;
  assert (_work > 0);
  _work--;

  if (_work == 0) {
    if (_cerr == 1)
      abort (wrap (mkref (this), &logw_impl::commit_abort, cb));
    else {
      _lptr->_verf = 0;
      _lptr->_lock = 0;
      _lptr->_commit = _commit_addr;
      _lptr->_seqn++;
      sync_lptr (wrap (mkref (this), &logw_impl::commit_sync_cb, cb));
    }
  }
}

bigint
logw_impl::make_sind (ptr<commit_record> c, cbc cb)
{
  assert (!_deltas.empty ());
  // make a single indirect block
  strbuf b;
  for (unsigned i=0; i<c->nhashes () && !_deltas.empty (); i++)
    sbigint (b, _deltas.pop_front ());
  str x = b;
  _work++;
  bigint sind = nsobj->insert_block
    (x.cstr (), x.len (), wrap (mkref (this), &logw_impl::commit_cb, cb));
  return sind;
}

void
logw_impl::commit (ptr<commit_record> c, cbc cb)
{
  assert (!_commit);
  assert (_work == 0);

  _commit = true;
  _cerr = 0;

  c->_head = _head;
  c->_prev = _pcommit;
  c->_ndeltas = _deltas.size ();

  // first, direct pointers
  for (unsigned i=0; i<c->nhashes () && !_deltas.empty (); i++)
    c->_deltas.push_back (_deltas.pop_front ());
  
  if (!_deltas.empty ()) // single indirect
    c->_sind = make_sind (c, cb);

  if (!_deltas.empty ()) { // double indirect
    strbuf b;
    while (!_deltas.empty ()) {
      bigint n = make_sind (c, cb);
      sbigint (b, n);
    }
    str x = b;
    _work++;
    c->_dind = nsobj->insert_block
      (x.cstr (), x.len (), wrap (mkref (this), &logw_impl::commit_cb, cb));
  }
  else
    c->_dind = 0;

  str data = c->serialize ();
  _work++;
  _commit_addr = nsobj->insert_block
    (data.cstr (), data.len (), wrap (mkref (this), &logw_impl::commit_cb, cb));
}

void
logw_impl::abort_cb (cbl cb, log_status status)
{
  reset ();
  cb (LOGERR_ABORTED);
}

void
logw_impl::abort (cbl cb)
{
  assert (_lptr);
  _lptr->_lock = 0;
  _lptr->_verf = 0;
  _lptr->_commit = _pcommit;
  _lptr->_seqn = _pseqn;
  _failed_stores = 0;
  sync_lptr (wrap (mkref (this), &logw_impl::abort_cb, cb));
}

