/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "rxx.h"

#include "ns.h"
#include "record.h"
#include "view.h"

int view_debug = 0;
int ns_debug = 0;
int cfs_debug = 0;

ptr<ns> nsobj = 0;
ptr<view> _view;
ptr<sfspriv> _sk;
static int _work = 0;
  
static str addkey = 0;
static str rule = 0;
static str name = "";
static bigint newkey;

void
store_cb (bool print, ns_stat status, bigint id)
{
  _work--;
  if (status != NS_OK)
    warn << "failed to store project block\n";
  else if (print) {
    char raw[20];
    mpz_get_rawmag_be (raw, sizeof (raw), &id);
    str xx = armor64A (raw, sizeof (raw));
    strbuf out;
    out << xx << "\n";
    out.tosuio ()->output (1);
  }
  if (_work == 0)
    exit (0);
}

void
new_view ()
{
  ref<view_constructor> v = New refcounted<view_constructor> ();
  _work++;
  v->store (_sk, wrap (store_cb, true));
}

void
modify_view (log_status status)
{
  if (status) {
    warn << "cannot load project block\n";
    exit (1);
  }

  ref<view_constructor> vc = New refcounted<view_constructor> (*_view);
  if (addkey) {
    if (!vc->add_entry (newkey, name)) {
      warn << "cannot add entry, duplicate name?\n";
      exit (1);
    }
  }
  if (rule)
    vc->set_notify (rule);
  _work++;
  vc->store (_sk, wrap (store_cb, true));
}

void
got_view (bool verbose, log_status status)
{
  if (status) {
    warn << "cannot load project block\n";
    exit (1);
  }

  strbuf out;
  str n = _view->notify ();

  if (verbose && n && n != "")
    out << "notify: " << n << "\n";
  
  for (unsigned i=0; i<_view->nentries (); i++) {
    if (verbose)
      out << "user: " << _view->name (i) << " ("
	  << armor64A (_view->head (i)) << ")\n";
    else {
      bigint n = _view->head (i);
      char raw[20];
      mpz_get_rawmag_be (raw, sizeof (raw), &n);
      str s = armor64A (raw, sizeof (raw));
      out << s << "\n";
    }
  }
  out.tosuio ()->output (1);
  exit (0);
}

void
usage ()
{
  warnx << "usage: " << progname << " projectid\n"
        << "         prints the public keys in the project\n";
  warnx << "       " << progname << " -s privkey\n"
        << "         creates a new project using <privkey>\n";
  warnx << "       " << progname << " -s privkey -r rule projectid\n"
        << "         adds <rule> as the new notification rule\n";
  warnx << "       " << progname << " -s privkey -n name -a key projectid\n"
        << "         adds <key> to project using <privkey>\n";
  warnx << "       " << progname << " -p projectid\n"
        << "         prints the public keys and Commit record hashes\n";
  exit (1);
}


void
start (bigint vid, bool verbose, bool newview)
{
  if (newview)
    new_view ();
  
  else if (addkey || rule) {
    if (addkey) {
      str keystr;

      rxx rx0 ("pk:([0-9a-zA-Z+-]*)$");
      rxx rx1 ("([0-9a-zA-Z+-]*)$");

      if (rx0.search (addkey))
        keystr = rx0 [1];
      else if (rx1.search (addkey))
        keystr = rx1 [1];
      else
        keystr = addkey;

      str dearm = dearmor64A(keystr);
      if (dearm)
        mpz_set_rawmag_be(&newkey, dearm.cstr(), dearm.len());
      else {
        warnx << "pastprojctl: key hash invalid\n";
        exit(1);
      }
    }
    _view = view::alloc (wrap (&modify_view), vid);
  }

  else
    _view = view::alloc (wrap (&got_view, verbose), vid);
}


int
main (int argc, char **argv)
{
  bool verbose = false;
  bool newview = false;
  char *skfile = 0;
  
  setprogname (argv[0]);
  random_init ();

  argv++;
  argc--;

  if (argc < 1)
    usage();

  while ( !strcmp(argv[0], "-r") || 
	  !strcmp(argv[0], "-n") ||
	  !strcmp(argv[0], "-a") ||
	  !strcmp(argv[0], "-s") ||
	  !strcmp(argv[0], "-p")) {

    if (!strcmp(argv[0], "-r")) {
      rule = str (argv[1], strlen (argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-n")) {
      name = str (argv[1], strlen (argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-s")) {
      skfile = argv[1];
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-a")) {
      addkey = str (argv[1], strlen(argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-p")) {
      verbose = true;
      argv += 1;
      argc -= 1;
    }
    if (argc == 0)
      break;
  }
  
  if (argc > 1 || (argc < 1 && skfile == NULL))
    usage();

  if (!addkey && (!name || name == "") && !rule && skfile)
    newview = true; 

  if ((newview && argc > 0) || (!newview && argc != 1)) {
    warn << "too many or too few args\n";
    usage ();
  }

  if ((newview || addkey || rule) && !skfile) {
    warn << "missing key\n";
    usage ();
  }

  if (addkey && name == "") {
    warn << "missing user name\n";
    usage ();
  }

  if (rule) {
    rxx rx ("^([CcEeRr]) ([a-zA-Z]*) ([\\w\\W\\s]*)$");
    if (!rx.search (rule)) {
      warn << "rule must be of the format: [CcEeRr] action arg\n";
      usage ();
    }
  }

  bigint vid = 0;

  if (argc == 1) {
    char sym[1024];
    str dearm;
    rxx rx ("project:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink(argv[0], sym, 1024);
    if (r > 0) sym[r] = '\0';
    if (rx.search (argv[0]) || (r > 0 && rx.search (sym)))
      dearm = dearmor64A(rx [1]);
    else
      dearm = dearmor64A(argv [0]);
    if (dearm) 
      mpz_set_rawmag_be(&vid, dearm.cstr(), dearm.len());
    else {
      warnx << "pastprojctl: invalid project ID\n";
      exit(1);
    }
  }

  if (newview || addkey || rule) {
    str key = file2wstr (skfile);
    if (!key) {
      warn << skfile << ": " << strerror (errno) << "\n";
      exit (1);
    }
    else if (!(_sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
      warn << "could not decode " << skfile << "\n";
      warn << key << "\n";
      exit (1);
    }
    if (vid == 0)
      vid = ns::pkid (_sk);
    else
      assert (vid == ns::pkid (_sk));
  }

  assert (vid != 0);
  nsobj = ns::alloc (vid);
  nsobj->connect (wrap (start, vid, verbose, newview));
  amain ();
}

