/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef RECORD_H
#define RECORD_H

#include "stdinc.h"
#include "pack.h"

class log_pointer
{
private:
  friend class logw_impl;
  friend class mklog;

  unsigned _seqn;
  bigint _commit;
  bigint _lock;
  bigint _verf;

protected:
  log_pointer ();

public:
  log_pointer (const log_pointer& u) { *this = u; }
  static ptr<log_pointer> alloc (str);

  void set_commit (bigint h) { _commit = h; }
  unsigned seqn () const { return _seqn; }
  bigint commit () const { return _commit; }
  bigint lock () const { return _lock; }
  bigint verf () const { return _verf; }

  str serialize() const;
};

typedef ptr<log_pointer> lptr;

struct commit_info
{
  bigint head;
  bigint rptr;
  bigint vptr;

  commit_info () { head = rptr = vptr = 0; }

  void serialize (strbuf &sb) const {
    sbigint (sb, head);
    sbigint (sb, rptr);
    sbigint (sb, vptr);
  }

  static bool parse (xin *in, commit_info &v)
  {
    if (!in->rbigint (v.head) ||
	!in->rbigint (v.rptr) ||
	!in->rbigint (v.vptr))
      return false;
    return true;
  }
};

class commit_record
{
private:
  friend class logw_impl;
  friend class mklog;

  bigint _head;
  bigint _prev;
  commit_info _depends;
  unsigned _time;

  unsigned _nhashes;
  unsigned _ndeltas;
  vec<bigint> _deltas;
  bigint _sind;
  bigint _dind;

  str _proposed_name;
  str _msg;
  str _btag;
  unsigned _flags;
  bigint _supersede;

protected:
  commit_record ();

public:
  commit_record (const commit_info& c, str n, str m,
                 str btag = 0, unsigned flags = 0, bigint supersede = 0, long commit_time = 0);
  str serialize () const;
  static ptr<commit_record> alloc (str);

  bigint head () const { return _head; }
  bigint prev () const { return _prev; }
  commit_info depends () const { return _depends; }
  str btag () const { return _btag; }
  unsigned ndeltas () const { return _ndeltas; }
  unsigned nhashes () const { return _nhashes; }
  const vec<bigint>& deltas () const { return _deltas; }
  bigint sind () const { return _sind; }
  bigint dind () const { return _dind; }

  unsigned time () const { return _time; }
  str proposed_name () const { return _proposed_name; }
  str msg () const { return _msg; }
  unsigned flags () const { return _flags; }
  bigint supersede () const { return _supersede; }

  static const int IGNORE_BRANCH = 0x1;
  static const int MERGE_WITH = 0x2;
  static const int BRANCH_OFF = 0x4;
};

typedef ptr<commit_record> crec;

class delta 
{
protected:
  friend class logw_impl;
  friend class mklog;

  bigint _prev;
  unsigned _rtype;

  void serialize_meta (strbuf &sb) const {
    sstr (sb, VERSION);
    suint (sb, _rtype);
  }
  bool parse_meta (xin *x);
  virtual const std::type_info& cxxtype () const { return typeid(*this); }

public:
  delta (int type = 0) : _rtype (type) {}
  virtual ~delta () {}
  virtual str serialize () const = 0;
  uint rtype () const { return _rtype; }

  template<class T> T* record () {
    if (typeid (T) != cxxtype () && typeid (refcounted<T>) != cxxtype ())
      return 0;
    return static_cast<T*> (this);
  }

  static bool parse_type (xin *x, unsigned &t);
};

typedef ptr<delta> deltap;

extern void get_deltas (crec c, callback<void,bigint>::ref ind_cb,
                             callback<void,bigint,bool>::ref cb);

#endif

