/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "dhash_common.h"
#include "stdinc.h"
#include "ns.h"
#include "view.h"
#include "record.h"

extern int view_debug;

class cpks_state;

class view_impl : public view {
private:
  bigint addr;
  vec<bigint> _head;
  vec<str> _name;
  str _notify;
  u_int64_t _fetcht;

  struct view_impl_state {
    ptr<view_state> s;
    unsigned nloaded;
    log_status error;
  };

  void take (ptr<view_constructor> c);
  void got_view (ptr<view_constructor> vc,
                 callback<void, log_status>::ref cb,
		 ptr<keyhash_payload> blk);
  
  void load_lptr (ptr<view_impl_state> vs, unsigned i,
                  callback<void, ptr<view_state>, log_status>::ref cb,
		  lptr u, long v, ns_stat status);
  
  void load_commit (ptr<view_impl_state> vs, unsigned i, bigint n,
                    callback<void, ptr<view_state>, log_status>::ref cb,
		    crec c, ns_stat status);
  
  void load_done (ptr<view_impl_state> v,
                  callback<void, ptr<view_state>, log_status>::ref cb);

  void checkpks_cb
    (ptr<view_impl_state> v, ptr<cpks_state> cs,
     callback<void, ptr<view_state>, log_status>::ref cb,
     vec<bigint> kv, ns_stat status);

public:
  view_impl (callback<void, log_status>::ref cb, bigint addr);
  view_impl (ptr<view_constructor> c);
  virtual ~view_impl () {}

  unsigned nentries () const { return _head.size(); }
  bigint head (unsigned i) const { return _head [i]; }
  str name (unsigned i) const { return _name [i]; }
  str notify () const { return _notify; }
  void fetch (const vervec&, const comvec&,
              callback<void, ptr<view_state>, log_status>::ref cb);
  void fetch (callback<void, ptr<view_state>, log_status>::ref cb);
};

// view constructor

view_constructor::view_constructor (const view &v)
{
  for (unsigned i = 0; i < v.nentries (); i++)
    add_entry_no_check (v.head (i), v.name (i));
  sort ();
  _notify = v.notify ();
}

bigint
view_constructor::store (ptr<sfspriv> sk,
                         callback<void, ns_stat, bigint>::ref cb)
{
  str data = serialize ();
  bigint n = ns::pkid (sk);
  nsobj->insert_pk (sk, data, cb);
  return n;
}

void
view_constructor::sort ()
  // slow but simple sort
{
  vec<bigint> h;
  vec<str> n;
  vec<bool> done;
  bool found = true;
  for (unsigned i=0; i<_head.size(); i++)
    done.push_back(false);

  while (found) {
    found = false;
    bigint sv;
    unsigned s = 0;
    for (unsigned i=0; i<_head.size(); i++) {
      if (!done[i]) {
	if (!found || _head[i] < sv) {
          found = true;
	  sv = _head[i];
	  s = i;
	}
      }
    }
    if (found) {
      h.push_back(_head[s]);
      n.push_back(_name[s]);
      done[s] = true;
    }
  }

  _head = h;
  _name = n;
}

str
view_constructor::serialize ()
{
  sort (); // paranoia
  strbuf sb;
  sstr (sb, VERSION);
  sint (sb, _head.size ());
  for (unsigned i=0; i<_head.size (); i++) {
    sbigint (sb, _head [i]);
    sbigint (sb, 0);
    sstr (sb, _name [i]);
  }
  sstr (sb, _notify);
  return sb;
}

void
view_constructor::set_notify (str x)
{
  _notify = x;
}

bool
view_constructor::add_entry (bigint h, str n)
{
  for (unsigned i=0; i<_head.size (); i++) {
    if (strncasecmp (_name [i].cstr (), n, n.len ()) == 0 &&
	_head [i] != h)
      return false;
  }
  for (unsigned i=0; i<_head.size (); i++) {
    if (_head [i] == h) {
      _name [i] = n;
      return true;
    }
  }
  _head.push_back (h);
  _name.push_back (n);
  sort ();
  return true;
}

void
view_constructor::add_entry_no_check (bigint h, str n)
{
  _head.push_back (h);
  _name.push_back (n);
}

bool
view_constructor::parse (xin *in)
{
  str swver;
  int n;
  if (!in->rstr (swver))
    return false;
  if (!in->rint (n))
    return false;
  for (int i=0; i<n; i++) {
    bigint h, b;
    str name;
    if (!in->rbigint (h) ||
	!in->rbigint (b) ||
	!in->rstr (name))
      return false;
    if (!name || name == "")
      warn << armor64A (h) << " has no user name\n";
    add_entry_no_check (h, name);
  }
  if (!in->rstr (_notify))
    return false;
  sort ();
  return true;
}

bool
view_constructor::parse_old (xin *in)
{
  int n;
  if (!in->rint (n))
    return false;
  for (int i=0; i<n; i++) {
    bigint h, b;
    if (!in->rbigint (h) ||
	!in->rbigint (b))
      return false;
    add_entry_no_check (h, "");
  }
  if (!in->rstr (_notify))
    return false;
  sort ();
  return true;
}

// view object

view_impl::view_impl (callback<void, log_status>::ref cb, bigint addr)
  : addr (addr)
{
  ptr<view_constructor> vc = New refcounted<view_constructor>;
  if (view_debug)
    _fetcht = getusec ();
  nsobj->retrieve_pk (addr, wrap (this, &view_impl::got_view, vc, cb));
}

ptr<view>
view::alloc (callback<void, log_status>::ref cb, bigint addr)
{
  return New refcounted<view_impl> (cb, addr);
}

void
view_impl::take (ptr<view_constructor> c)
{
  assert (c);
  for (unsigned i = 0; i < c->_head.size (); i++) {
    _head.push_back (c->_head[i]);
    _name.push_back (c->_name[i]);
  }
  _notify = c->_notify;
}

void
view_impl::got_view (ptr<view_constructor> vc, 
                     callback<void, log_status>::ref cb,
		     ptr<keyhash_payload> p)
{
  if (p == 0) {
    warn << "could not fetch project block\n";
    cb (LOGERR_NOVIEW);
    return;
  }

  if (view_debug) {
    u_int64_t t = getusec () - _fetcht;
    warn << "fetchproj: reply, +" << (long) t << "\n";
  }

  xin in (p->buf ());
  if (!vc->parse (&in)) {
    xin in2 (p->buf ());
    if (!vc->parse_old (&in2)) {
      cb (LOGERR_NOVIEW);
      return;
    }
  }
  take (vc);
  cb (LOG_OK);
}

struct cpks_state {
  vec<bigint> missing;
  vec<unsigned> missing_idx;
  vec<bigint> check;
  vec<unsigned> check_idx;
};

void
view_impl::checkpks_cb
  (ptr<view_impl_state> v, ptr<cpks_state> cs,
   callback<void, ptr<view_state>, log_status>::ref cb,
   vec<bigint> kv, ns_stat status)
{
  if (view_debug) {
    u_int64_t t = getusec () - _fetcht;
    warn << "checkpks: reply, +" << (long) t << "\n";
  }

  if (status) {
    if (view_debug)
      warn << "checkpks failed, try normal fetch, could take awhile\n";
    fetch (cb);
    return;
  }

  else {
    // count the keys that have not changed
    v->nloaded += (cs->check.size () - kv.size ());

    for (unsigned i=0; i<kv.size (); i++) {
      // warn << armor64A (kv [i]) << " has changed\n";
      unsigned j=0;
      for (j=0; j<cs->check.size (); j++) {
	if (kv [i] == cs->check [j])
	  break;
      }
      if (j < cs->check.size ()) { // we asked the server
        nsobj->retrieve_lptr
	  (kv [i],
	   wrap (this, &view_impl::load_lptr, v, cs->check_idx [j], cb));
      }
      else // it was not asked, just assume it's okay?
	v->nloaded++;
    }

    load_done (v, cb);
  }
}

void
view_impl::fetch (const vervec& pkv, const comvec& v,
                  callback<void, ptr<view_state>, log_status>::ref cb)
{
  if (nentries () < 2) {
    fetch (cb);
    return;
  }

  vec<long> lv;
  vec<bigint> missing;

  ptr<cpks_state> cs = New refcounted<cpks_state>;
  unsigned i, j;

  ptr<view_impl_state> vs = New refcounted<view_impl_state> ();
  vs->s = New refcounted<view_state> ();

  for (i=0, j=0; i<_head.size ();) {
    if (j < pkv.size ()) {
      if (_head [i] == pkv.head (j)) {
        lv.push_back ((long)pkv.seqn (j));
	cs->check.push_back (_head [i]);
	cs->check_idx.push_back (i);
        vs->s->lptrs.push_back (0);
	i++;
	j++;
      }
      else { // we assume that pkv is missing this element
	assert (pkv.head (j) > _head [i]);
	cs->missing.push_back (_head [i]);
	cs->missing_idx.push_back (i);
        vs->s->lptrs.push_back (0);
	i++;
      }
    }
    else { // _head [i] not in pkv
      cs->missing.push_back (_head [i]);
      cs->missing_idx.push_back (i);
      vs->s->lptrs.push_back (0);
      i++;
    }
  }

  if (j != pkv.size ())
    warn << "ignoring some pub keys!\n";

  vs->s->pkv = pkv;
  vs->s->commits = v;

  if (view_debug)
    warn << "checking " << cs->check.size () << " keys, "
         << "fetching " << cs->missing.size () << " keys.\n";

  vs->error = LOG_OK;
  vs->nloaded = 0;

  if (view_debug)
    _fetcht = getusec ();

  for (unsigned i=0; i<cs->missing.size (); i++) {
    if (view_debug)
      warn << armor64A (cs->missing [i]) << " not checked\n";
    nsobj->retrieve_lptr
      (cs->missing [i],
       wrap (this, &view_impl::load_lptr, vs, cs->missing_idx [i], cb));
  }

  if (cs->check.size () > 0)
    nsobj->checkpks
      (cs->check, lv, wrap (this, &view_impl::checkpks_cb, vs, cs, cb));
  else
    load_done (vs, cb);
}

void
view_impl::fetch (callback<void, ptr<view_state>, log_status>::ref cb)
{
  ptr<view_impl_state> v = New refcounted<view_impl_state> ();
  v->s = New refcounted<view_state> ();

  v->error = LOG_OK;
  v->nloaded = 0;

  if (view_debug)
    _fetcht = getusec ();

  for (unsigned i=0; i<_head.size (); i++) {
    v->s->lptrs.push_back (0);
    nsobj->retrieve_lptr
      (_head [i], wrap (this, &view_impl::load_lptr, v, i, cb));
  }
  load_done (v, cb);
}

void
view_impl::load_lptr (ptr<view_impl_state> v, unsigned i,
                      callback<void, ptr<view_state>, log_status>::ref cb,
		      lptr u, long pkv, ns_stat status)
{
  if (!u) {
    v->nloaded++;
    v->error = LOGERR_NOCPTR;
    warn << "cannot load lptr " << armor64A (_head [i]) << "\n";
    load_done (v, cb);
  }
  else {
    if (view_debug) {
      u_int64_t t = getusec () - _fetcht;
      warn << "load_lptr: reply, +" << (long) t << "\n";
    }

    v->s->lptrs [i] = u;
    v->s->pkv.add (_head [i], (unsigned)pkv);

    if (v->s->commits.size () > i &&
	v->s->commits.head (i) == _head [i] &&
	v->s->commits.rptr (i) == u->commit ()) {
      if (view_debug)
	warn << "commit record pointer has not changed\n";
      v->nloaded++;
      load_done (v, cb);
    }
    else if (u->commit () == 0) {
      v->nloaded++;
      v->s->commits.add (_head [i], 0);
      load_done (v, cb);
    }
    else
      nsobj->retrieve_commit
	(u->commit (),
	 wrap (this, &view_impl::load_commit, v, i, u->commit (), cb));
  }
}

void
view_impl::load_commit (ptr<view_impl_state> v, unsigned i, bigint n,
                        callback<void, ptr<view_state>, log_status>::ref cb,
			crec c, ns_stat status)
{
  if (!c) {
    warn << "cannot load Commit record " << armor64A (n) << "\n";
    warn << "for user " << armor64A (_head [i]) << "\n";
    v->nloaded++;
    v->error = LOGERR_NOCOMMIT;
  }
  else {
    v->nloaded++;
    v->s->commits.add (_head [i], n);
  }

  load_done (v, cb);
}

void
view_impl::load_done (ptr<view_impl_state> v,
                      callback<void, ptr<view_state>, log_status>::ref cb)
{
  assert (v->nloaded <= _head.size ());
  if (v->nloaded != _head.size ())
    return;
  cb (v->s, v->error);
}

