/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include "pastwatch.h"
#include "update.h"
#include "opts.h"
#include "path.h"
#include "pdb.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;
extern int update_debug;
extern str get_tag_of_commit (bigint cn);

extern void cfsr_lookup_v (bigint dir, str path, bigint s, cbcfsr cb);
extern void checkout_one (bigint, str, str, bigint, cbi cb);
extern int  check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void check_skip   (str dirpath, str name, bool& ignore_entry);
extern int  check_ignore (str dirpath, str name, bool& ignore);
extern void read_file    (int, bigint, bigint,
                          callback<void, cfsrw_status>::ref);
extern int  isdiff       (const char *f1, const char *f2);
extern int  isconf       (const char *path);
extern int  merge        (const char *fb, const char *tb,
                          const char *fw, const char *tfw,
                          const char *fr, const char *tfr, bool check_only);

#define ISDIR(sb) ((sb.st_mode & S_IFDIR) == S_IFDIR)
#define ISREG(sb) ((sb.st_mode & S_IFREG) == S_IFREG)
#define ISLNK(sb) ((sb.st_mode & S_IFLNK) == S_IFLNK)

class update_object : public virtual refcount, public update_object_spec
{
private:
  struct ude_t {
    str name;
    str path;
    str cfspath;
    cfs_ftype type;
    bigint fh;
    bigint dirfh;

    pde_t get_pde () {
      pde_t e;
      e.name = name;
      e.path = path;
      e.cfspath = cfspath;
      e.mode = mode;
      e.type = type;
      e.fh = fh;
      e.dirfh = dirfh;
      return e;
    }

    mtime_t mtime;
    mtime_t db_mtime;
    unsigned mode;
    bigint saddr;
    bigint sticky;
    bigint newsaddr;
    // below are attributes from repository
    bigint commit;
    bigint mode_commit;
    unsigned revision;
    bool dirempty;
    
    char status; 
    // I'm not totally sure what these values mean, so these are guesses:
    // '?' : not set yet.
    // ' ' : file exists in the working copy and the CFS repository (marked on CFS entry)
    // '-' : file exists under a different name (marked on local file entry)
    //     : or file is no longer relevant in this directory (marked on local file)
    // 'n' : file exists in CFS repository (marked on local file entry)
    // 'A' : file should be added to CFS repository (marked on local file entry)
    // 'R' : file should be removed from CFS repository (marked on local file entry)

    bool renamed;
    bool conflict;
    bool compare;
    bool modified;

    unsigned rnlink; // only used for rescued files
    cfs_ftype rtype; // only used for rescued files
  };
  typedef vec<ude_t> udev_t;

  struct upst_t {
    str path;
    str cfspath;
    str filter;
    bigint dirfh;
    udev_t local;
    udev_t cfs;
  };
  typedef ptr<upst_t> upstp_t;

  int work;
  int error;
  cbi final_cb;
  
  bool checkonly;
  bool updatedir;

  vec<str> dirs;
  vec<str> filters;
  vec<str> dirs_to_remove;

  bigint rep_root;
  bigint rep_saddr;

  void
  update_done (int err)
  {
    if (err)
      error = err;
    if (work == 0) {
      for (unsigned i=0; i<dirs_to_remove.size (); i++)
        if (pmgr->remove_dir (dirs_to_remove [i]))
	  warn << "cannot remove directory " << dirs_to_remove [i] 
	       << ": directory may be non-empty!\n";
      final_cb (error);
    }
  }

  void
  check_status (str path)
  {
    struct dirops dpv;
    str n;
    str d = get_parent_dir (path, n);
    if (!get_dirops (d, dpv)) {
      for (unsigned i=0; i<dpv.add.size (); i++) {
	if (dpv.add [i]->name == n)
	  return;
      }
      for (unsigned i=0; i<dpv.remove.size (); i++) {
	if (dpv.remove [i]->name == n)
	  return;
      }
    }
    pastwatch_status ("?", path);
  }

  void
  update_next_dir (cbi cb)
  {
    if (dirs.empty ())
      cb (0);
    else {
      str path = dirs.pop_front ();
      str filter = filters.pop_front ();

      struct stat sb;
      int r = lstat (path, &sb);
      if (r == 0) {
        str cfspath;
        if (!get_cfspath (path, cfspath))
          update_directory (path, cfspath, filter, cb);
        else {
          if (r == 0 && !pmgr->exists (path)) {
	    bool ignore = false;
            str n;
	    str d = get_parent_dir (path, n);
            check_ignore (d, n, ignore);
            if (!ignore)
	      check_status (path);
          }
	  update_next_dir (cb);
        }
      }
      else
	update_next_dir (cb);
    }
  }

  void
  op_done (upstp_t upstp, cbi cb, int err)
  {
    work --;
    if (err || error) {
      error = -1;
      if (work == 0) {
        cb (error);
      }
      return;
    }
    if (work == 0) {
      // set attribute for the whole dir if the dir changed
      if (upstp->filter == "") {
        ptr<pfe> x = pmgr->lookup_in_parent_dir (upstp->path);
        if (x) {
	  if (arg_resetsticky)
	    x->sticky = 0;
	  if (x->sticky == 0) {
	    x->saddr = rep_saddr;
	    if (arg_revision0 || arg_tag0 != "")
	      x->sticky = rep_saddr;
	    else
	      x->sticky = 0;
	  }
	  if (!arg_noupdate) {
	    pmgr->schedule_add (upstp->path, *x);
	  }
	  if (arg_command == cmd_merge) {
	    assert (arg_tag0 != "");
            bigint n;
            dearmor64A_bigint (arg_tag0, n);
	    pmgr->schedule_supersedes (upstp->path, n);
	  }
        }
      }
      update_next_dir (cb);
    }
  }

  void
  compare_file_have_base (str base, int tfd, str current,
                          ude_t *de, cbi cb, cfsrw_status status)
  {
    close (tfd);
    if (status) {
      unlink (current);
      unlink (base);
      cb (-1);
      return;
    }
  
    bool conflict = false;
    if (!checkonly)
      pastwatch_status ("Merging changes into", de->path);

    int c = merge (base.cstr (), "base",
                   de->path.cstr (), "working",
		   current.cstr (), "current", checkonly);
    if (c < 0) {
      warn << de->path << ": cannot merge changes from repository\n";
      unlink (current);
      unlink (base);
      cb (-1);
      return;
    }
    else if (c)
      conflict = true;
  
    if (!conflict) {
      int r = isdiff (de->path.cstr (), current.cstr ());
      if (r < 0) {
        warn << de->path << ": cannot compare against repository\n";
        unlink (current);
        unlink (base);
        cb (-1);
        return;
      }
      if (r) {
	if (de->sticky == 0)
	  pastwatch_status ("M", de->path);
	else {
	  strbuf b;
	  b << "different from " << get_tag_of_commit (de->commit);
	  pastwatch_status ("M", de->path, b);
	  requires_update = true;
	}
        to_commit.push_back (de->get_pde ());
        if (!arg_noupdate && !checkonly && de->status == ' ') {
          // file changed, preserve old mtime until commit
	  pmgr->schedule_add
	    (de->path, pfe (de->fh, de->newsaddr, de->db_mtime, de->sticky));
	}
      }
      else { // either content did not change or merge downloaded the changes
        pastwatch_status ("P", de->path);
        if (!arg_noupdate && !checkonly && de->status == ' '){
	  pmgr->schedule_add
	    (de->path, pfe (de->fh, de->newsaddr, de->mtime, de->sticky));
	}
      }
    }
    else {
      has_conflict = true;
      pastwatch_status ("C", de->path);
      if (!arg_noupdate && !checkonly && de->status == ' ') {
	// conflict, but changes are merged.  this means we can store
	// the new version vector from repository, but leave the old
	// mtime so that we know we still need to compare the file
	// again on the next update.
        pmgr->schedule_add
	  (de->path, pfe (de->fh, de->newsaddr, de->db_mtime, de->sticky));
      }
    }
  
    unlink (current);
    unlink (base);
    if (conflict)
      requires_update = true;
    cb (0);
  }

  void
  compare_file_have_repository (str tfn, int tfd, ude_t *de, ude_t *cde,
                                cbi cb, cfsrw_status status)
  {
    close (tfd);
    if (status) {
      unlink (tfn);
      cb (-1);
      return;
    }

    int r = isdiff (de->path.cstr (), tfn.cstr ());
    if (r < 0) {
      warn << de->path << ": cannot compare against repository\n";
      unlink (tfn);
      cb (-1);
      return;
    }
    if (r) {
      if (de->commit != cde->commit && de->saddr != 0) {
	// we need to find the base version for merging... this is the
	// version that caused the fork between my working copy of the
	// file and the copy of the file in the repository.
        ptr<btree_srv> b = _cfsrw->getreplica ()->btree ();
	bigint j;
	if (!b->find_join (de->commit, cde->commit, j)) {
          warn << "cannot find a base version to use for merging\n";
	  warn << "using the checked-out version instead\n";
	  warn << "this may result in extra conflicts\n";
          j = de->saddr;
	}
	else {
	  if (j == cde->commit || j == de->commit) {
	    // from the same branch, we just want to use our
	    // checked-out version
	    j = de->saddr;
	  }
	  else {
	    btree_node x;
	    if (!b->lookup_by_commit (j, x)) {
	      assert (0);
              j = de->saddr;
	    }
	    else
	      j = x.commit.vptr;
	  }
	}

        str s = pastwatch_tmpfile ();
        int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
        read_file
          (fd, de->fh, j,
           wrap (mkref (this), &update_object::compare_file_have_base,
	         s, fd, tfn, de, cb));
      }
      else {
	if (de->sticky == 0)
	  pastwatch_status ("M", de->path);
	else {
	  strbuf b;
	  b << "different from " << get_tag_of_commit (de->commit);
	  pastwatch_status ("M", de->path, b);
	  requires_update = true;
	}

        to_commit.push_back (de->get_pde ());
        if (!arg_noupdate && !checkonly && de->status == ' '){
          // file changed, preserve old mtime until commit
	  pmgr->schedule_add
	    (de->path, pfe (de->fh, de->newsaddr, de->db_mtime, de->sticky));
	}
	unlink (tfn);
	cb (0);
      }
    }

    else {
      if (!arg_noupdate && de->status == ' ') {
        if (update_debug)
	  warn << de->path << ": same, no comparison\n";
	if (de->sticky == 0 &&
	    de->newsaddr == my_branch_head_snapshot (_cfsrw->getreplica ())){
	  // file should have the saddr of the branch head... don't
	  // insert that saddr into pmgr yet: if this is a commit
	  // operation and a new snapshot is built, use that
	  // snapshot's address. otherwise, driver.C will insert the
	  // branch head saddr later.
          pmgr->schedule_add (de->path, pfe (de->fh));
	}else if (!checkonly){
          pmgr->schedule_add (de->path, pfe (de->fh, de->newsaddr, de->sticky));
	}
      }
      unlink (tfn);
      cb (0);
    }
  }
  
  void
  compare_file (ude_t *de, ude_t *cde, cbi cb)
  {
    int r = isconf (de->path);
    if (r < 0) {
      warn << "cannot open file " << de->path << "\n";
      cb (-1);
      return;
    }
    if (r) {
      pastwatch_status ("C", de->path);
      if (!arg_noupdate && !checkonly && de->status == ' ') {
        // still conflict, don't merge.  keep old version and mtime.
        pmgr->schedule_add
	  (de->path, pfe (de->fh, de->saddr, de->db_mtime, de->sticky));
      }
      has_conflict = true;
      cb (0);
      return;
    }
  
    str s = pastwatch_tmpfile ();
    int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
    read_file
      (fd, de->fh, de->newsaddr,
       wrap (mkref (this), &update_object::compare_file_have_repository,
	     s, fd, de, cde, cb));
  }
  
  void
  readlink_cb (ude_t* de, ude_t *cde, cbi cb, cfsrcallp c, cfsrw_status status)
  {
    char buf [1024];
    int r = readlink (de->path, buf, 1024);

    if (status || r < 0) {
      warn << de->path << ": cannot read link value from cfs or local dir.\n";
      cb (-1);
      return;
    }

    str l (buf, r);
    cfsr_readlink_res *res = static_cast<cfsr_readlink_res *> (c->res ());
  
    if (l != res->dest) {
      if (!de->modified) {  // just download the link from repository
        if (!checkonly) {
          r = unlink (de->path);
          if (r < 0) {
            warn << de->path << ": cannot remove to create new symlink.\n";
            cb (-1);
            return;
          }
          r = symlink (res->dest, de->path);
          if (r < 0) {
            warn << de->path << ": cannot create symbolic link.\n";
            cb (-1);
            return;
          }
	  if (!arg_noupdate && de->status == ' ') {
	    pmgr->schedule_add
	      (de->path, pfe (de->fh, de->newsaddr, de->sticky));
	  }
        }
        pastwatch_status ("S", de->path);
      }
      else { // modified
        if (de->commit != cde->commit) {
	  // changed in repository

	  pastwatch_status ("C", de->path);
	  has_conflict = true;
	  de->conflict = true;

	  // add the old version to preserve conflict until it is
	  // resolved. preserve mtime until commit.
          if (!arg_noupdate && !checkonly && de->status == ' '){
	    pmgr->schedule_add
	      (de->path, pfe (de->fh, de->saddr, de->db_mtime, de->sticky));
	  }
        }
        else {
	  if (de->sticky == 0)
	    pastwatch_status ("M", de->path);
	  else {
	    strbuf b;
	    b << "different from " << get_tag_of_commit (de->commit);
	    pastwatch_status ("M", de->path, b);
	    requires_update = true;
	  }
          to_commit.push_back (de->get_pde ());
          if (!arg_noupdate && !checkonly && de->status == ' '){
	    // preserve mtime until commit.
	    pmgr->schedule_add
	      (de->path, pfe (de->fh, de->newsaddr, de->db_mtime, de->sticky));
	  }
        }
      }
    }
    else if (!arg_noupdate && !checkonly && de->status == ' '){
      pmgr->schedule_add (de->path, pfe (de->fh, de->newsaddr, de->sticky));
    }

    cb (0);
  }

  void
  compare (ude_t* de, ude_t* cde, cbi cb)
  {
    assert (de->fh == cde->fh);

    // compare the version represented by de with that by cde.  if
    // sticky flag is on, compare de against the sticky version,
    // rather than cde (by default, cde is the version in the branch
    // head). in this case, we don't want to set cde->commit to
    // de->commit, as de->commit may not be the same as the sticky
    // version.
    
    if (de->sticky != 0) {
      cfsrcallp c = New refcounted<cfsrcall>
	(CFSRPROC_GETATTR,
	 wrap (mkref (this),
	       &update_object::compare_check_sticky, de, cde, cb));
      cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
      arg->object = de->fh;
      cfsr_call (de->sticky, c);
    }
    else
      compare_sticky_ready (de, cde, cb);
  }

  void
  compare_check_sticky (ude_t* de, ude_t* cde, cbi cb,
                        cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      cb (-1);
      return;
    }

    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (c->res ());
    cde->commit = res->attr->commit;
    cde->type = res->attr->type;
    cde->mode = res->attr->mode;
    cde->revision = res->attr->revision;
    compare_sticky_ready (de, cde, cb);
  }

  void
  compare_sticky_ready (ude_t* de, ude_t* cde, cbi cb)
  {
    if (de->type == CFSLNK || de->type == CFSREG) {

      // first, look to see if the mode has changed

      bool mode_changed = false;
      if (de->mode != 0 && de->mode != cde->mode) {
        if (de->mode_commit != cde->mode_commit) {
	  // repository copy of the file has a different mode
          pastwatch_status ("p", de->path, "reset permission");
          chmod (de->path, cde->mode);
	}
	else {
	  // working version's mode has been changed
	  mode_changed = true;

          if (de->sticky == 0)
            pastwatch_status ("m", de->path);
          else {
	    strbuf b;
	    b << "different from " << get_tag_of_commit (de->commit);
	    pastwatch_status ("m", de->path, b);
            requires_update = true;
          }

          to_setattr.push_back (de->get_pde ());
          if (!arg_noupdate && !checkonly && de->status == ' '){
            // file changed, preserve old mtime until commit
	    pmgr->schedule_add
	      (de->path, pfe (de->fh, de->newsaddr, de->db_mtime, de->sticky));
	  }
	}
      }

      // see if the file changed in repository (i.e. someone checked
      // in some changes we don't yet know). if the file is to be
      // re-linked (i.e. saddr == 0), we just compare against the
      // latest file in the repository.

      if (de->commit != cde->commit || de->modified) {
        if (update_debug)
	  warn << de->path << ": changed in repository or working dir\n";
	if (de->type == CFSLNK) {
          cfsrcallp c = New refcounted<cfsrcall> 
	    (CFSRPROC_READLINK,
	     wrap (mkref (this), &update_object::readlink_cb, de, cde, cb));
          cfsr_readlink_arg *arg = static_cast<cfsr_readlink_arg *> (c->arg ());
          arg->file = de->fh;
	  // newsaddr is basically rep_saddr, unless sticky bit is on
          cfsr_call (de->newsaddr, c); 
	}
	else if (checkonly && de->renamed) {
	  cb(0);
	} else {
          compare_file (de, cde, cb);
	}
      }

      else {
        if (!mode_changed && !arg_noupdate && de->status == ' ') {
          if (update_debug)
	      warn << de->path << ": same, no comparison\n";
	  if (!checkonly) {
	    if (de->sticky == 0 &&
	        de->newsaddr ==	my_branch_head_snapshot (_cfsrw->getreplica ())){
	      // for this file, pmgr should have the saddr of the
	      // branch head.  don't insert that saddr into pmgr yet:
	      // if this is a commit operation and a new snapshot is
	      // built, driver.C will pass the new snapshot's address
	      // into pmgr later.  otherwise, driver.C will pass the
	      // branch head's saddr to pmgr.
	      pmgr->schedule_add (de->path, pfe (de->fh));
	    } else {
	      pmgr->schedule_add
	        (de->path, pfe (de->fh, de->newsaddr, de->mtime, de->sticky));
	    }
	  }
	}
	cb (0);
      }
    }
    else { 
      // don't need to handle directory entry here; the entry was
      // pushed onto the dirs vector
      if (!arg_noupdate && !checkonly && de->status == ' ') {
        pmgr->schedule_add (de->path, pfe (de->fh, de->newsaddr, de->sticky));
      }
      cb (0);
    }
  }

  // compare_directory () synchronizes working directory against the
  // corresponding directory in the repository.

  void
  compare_directory (upstp_t upstp, cbi cb)
  {
    if (upstp->local.size () == 0 && upstp->cfs.size () == 0) {
      if (upstp->filter != "") {
	warn << "nothing known about " << upstp->filter << "\n";
	cb (0);
	return;
      }
    }

    if (upstp->filter == "")
      warn << "updating " << upstp->path << "\n";
    else
      warn << "updating " << upstp->filter << "\n";

    vec <str> local_remove;
    vec <str> local_rename_old;
    vec <str> local_rename_new;
    vec <str> local_rename_cfs;
    vec <int> local_add;

    // go thru files in the working directory to look for conflicts
    for (unsigned i=0; i<upstp->local.size (); i++) {
      ude_t *de = &(upstp->local [i]);
      de->conflict = false;
      de->compare = false;
      str path = de->path;  // local file name

      bool badtype = false;
      bool fh_exists = false;
      bool name_exists = false;
      bool changed = false;
      bool dempty = false;
      str rep_name;

      // check if file name and fh exist in repository; and if fh does
      // exist, what the name of the file is

      for (unsigned j=0; j<upstp->cfs.size (); j++) {
        ude_t *cde = &(upstp->cfs [j]);
	// skip fake entries that represent new files in the
	// repository; these files have not been committed.
	if (cde->status != 'n' && cde->status != ' ')
	  continue;
        if (de->fh != 0 && cde->fh == de->fh) { // fh exists
	  cde->status = ' ';
          if (de->type != cde->type && de->status == ' ') {
	    badtype = true;
	    break;
          }
	  else {
	    if (de->commit != cde->commit)
	      changed = true;
	    if (cde->type == CFSDIR && cde->dirempty)
	      dempty = true;
            fh_exists = true;
	    rep_name = cde->name;
	  }
        }
        if (cde->name == de->name)
	  // Assuming the repository is the authority, true means
	  // the file's name "exists" in the local directory.
          name_exists = true; 
      }

      if (badtype) {
	pastwatch_status
	  ("C", path, "file type does not match one in repository");
	has_conflict = true;
	de->conflict = true;
      }

      else if (de->status == 'A' && de->fh != 0) {
	if (de->rnlink != 0) {
	  pastwatch_status
	   ("C", armor64A (de->fh), " already linked in repository");
	  has_conflict = true;
	  de->conflict = true;
	}
	else if (de->rtype != CFSREG) {
	  pastwatch_status
	   ("C", armor64A (de->fh),
	    " is a directory or symlink, cannot relink");
	  has_conflict = true;
	  de->conflict = true;
	}
      }

      else if (fh_exists) {

	if (de->status == 'A') {
	  pastwatch_status
	    ("C", path, "new file exists in repository");
	  has_conflict = true;
	  de->conflict = true;
	}
	else if (changed && de->status == 'R') {
	  pastwatch_status
	    ("C", path, "changed in the repository, will not remove.");
	  has_conflict = true;
	  de->conflict = true;
	}
	else if (de->type == CFSDIR && de->status == 'R' && !dempty) {
	  pastwatch_status
	    ("C", path, "directory not empty in repository.");
	  has_conflict = true;
	  de->conflict = true;
	}
        else if (de->status == 'R' && de->modified) {
	  pastwatch_status
	    ("C", path, "should be removed and is still here");
	  has_conflict = true;
	  de->conflict = true;
	}
	else if (rep_name == de->name) {
	  if (update_debug)
	    warn << path << ": same fh, same name\n";
        }
        else { // fh exists, but under a different name
	  if (update_debug)
	    warn << path << ": same fh, different name.  Repository name is: " 
		 << rep_name << "\n";
	  if (de->status == 'R') {
	    pastwatch_status
	      ("C", path, "file removed in repository");
	    has_conflict = true;
	    de->conflict = true;
	  }
	  else {
	    // first, remember old name
	    de->renamed = true;
	    local_rename_old.push_back (de->name);
	    local_rename_new.push_back (rep_name);
	    de->name = rep_name;
            de->path = make_path (upstp->path, rep_name);
            de->cfspath = make_path (upstp->cfspath, rep_name);
	    local_rename_cfs.push_back (de->cfspath);
	  }
        }
      }

      else if (name_exists && de->status == 'A') {
	pastwatch_status
	  ("C", path, "created independently by second party.");
	has_conflict = true;
	de->conflict = true;
      }

      else if (!de->modified) {
        // not modified, fh no longer exists in the same directory
        if (update_debug)
	  warn << path << ": not modified, fh removed from dir\n";
        if (name_exists) {
	  pastwatch_status
	    ("", path, "appears in repository as a different file.");
	  pastwatch_status
	    ("", path, "may need to run update again to get new file.");
          local_remove.push_back (de->name);
	  de->status = '-';
        }
        else if (de->status != 'A' && de->status != '?') {
	  pastwatch_status ("", path, "no longer relevant in this directory.");
          local_remove.push_back (de->name);
	  de->status = '-';
        }
      }

      // !fh_exists, modified
      else if (de->status != '?' && de->status != 'A') {
        if (update_debug)
	  warn << path << ": modified, fh removed from dir\n";
	pastwatch_status
	  ("C", path, "modified, but no longer relevant in this directory.");
	has_conflict = true;
	de->conflict = true;
      }

      if (de->status == '?') {
	if (name_exists) {
	  pastwatch_status
	    ("C", path, "someone else created this file in the repository already.");
	  has_conflict = true;
	  de->conflict = true;
	}
	else {
	  bool found = false;
	  for (unsigned i=0; i<dirs.size (); i++) {
	    if (dirs [i] == path)
	      found = true;
	  }
	  if (!found) {
	    bool ignore = false;
            check_ignore (upstp->path, de->name, ignore);
            if (!ignore)
	      pastwatch_status ("?", path);
	  }
	}
      }

      else if (de->status == 'A') {
	if (de->fh == 0)
	  pastwatch_status ("A", path);
	else
	  pastwatch_status ("L", path);
	to_add.push_back (de->get_pde ());
      }

      else if (de->status == 'R') {
        pastwatch_status ("R", path);
	to_remove.push_back (de->get_pde ());
      }
    }

    for (unsigned i=0; i<upstp->cfs.size (); i++) {
      ude_t *cde = &(upstp->cfs [i]);
      if (cde->status == 'n') {
	// pastwatch_status ("new", cde->path);
	local_add.push_back (i);
      }
    }

    // if there is a conflict, don't continue, because we may not be
    // able to download or remove/rename some of the files...
    if (has_conflict) {
      warn << "please resolve above conflicts first!\n";
      cb (0);
      return;
    }

    if (local_remove.size () || local_rename_old.size () || local_add.size ())
      requires_update = true;

    if (!checkonly) { // update the local working directory

      // do remove before rename, so rename won't fail
      for (unsigned i=0; i<local_remove.size (); i++) {
        str path = make_path (upstp->path, local_remove [i]);
        struct stat sb;
        if (lstat (path, &sb)) {
          strbuf b;
          b << "cannot remove " << path;
          perror (str (b).cstr ());
          cb (-1);
          return;
	}
	if (ISDIR (sb)) {
	  dirs_to_remove.push_back (path);
	}
	else if (unlink (path)) {
          strbuf b;
          b << "cannot remove " << path;
          perror (str (b).cstr ());
          cb (-1);
          return;
        }
	if (!arg_noupdate)
	  pmgr->schedule_remove (path);
      }
      local_remove.clear ();
    
      // rename. do this in multiple passes.  in each pass, if target
      // exists, skip, and do it again on a later pass.  the idea is
      // that we don't remove a target, with rename, that needs to be
      // renamed later.
    
      vec <bool> local_rename_done;
      local_rename_done.setsize (local_rename_old.size ());
      for (unsigned i=0; i<local_rename_done.size (); i++)
        local_rename_done [i] = false;
    
      bool changed = true;
      while (changed) {
        changed = false;
        for (unsigned i=0; i<local_rename_old.size (); i++) {
          str oldf = make_path (upstp->path, local_rename_old [i]);
          str newf = make_path (upstp->path, local_rename_new [i]);
	  str cfsp = local_rename_cfs [i];
          struct stat sb;
          int r = lstat (newf, &sb);
          if (r < 0 && errno == ENOENT) {
	    // target does not exist, can rename
	    changed = true;
	    if (rename (oldf, newf)) {
	      strbuf b;
	      b << "cannot rename " << oldf << " to " << newf;
	      perror (str (b).cstr ());
	      cb (-1);
	      return;
	    }
	    pastwatch_status
	      ("Rename", local_rename_old [i], local_rename_new [i]);

	    r = lstat (newf, &sb);
	    if (r == 0 && (ISDIR(sb))){
	      dirs.push_back (newf);
	      filters.push_back ("");
	      pmgr->schedule_rename (newf, cfsp);
	    }

	    local_remove.push_back (oldf);
	    local_rename_done [i] = true;
	    // don't need to pmgr->schedule_add newf; all renamed
	    // entries have the target entry marked in the
	    // cfs_handle_in_local vector, so the code that pairwise
	    // compare files will do the pmgr->schedule_add.
          }
          else if (r < 0) {
	    strbuf b;
	    b << "cannot stat(1) " << newf;
	    perror (str (b).cstr ());
	    warn << "error: " << errno << " " << ENOENT << "\n";
	    cb (-1);
	    return;
          }
        }
      }

      for (unsigned i=0; i<local_remove.size (); i++) {
        str s = local_remove [i];
        assert (s != "");
        str path = make_path (upstp->path, s);
	if (!arg_noupdate)
          pmgr->schedule_remove (path);
      }
  
      for (unsigned i=0; i<local_rename_done.size (); i++) {
        if (!local_rename_done [i]) {
          warn << "rename " << local_rename_old [i] << " -> "
	       << local_rename_new [i] << " skipped!\n";
        }
      }
    } else {
      // notify the user of renames even if checkonly == true
      for (unsigned i=0; i<local_rename_old.size (); i++) {
	pastwatch_status
	  ("Rename", local_rename_old [i], local_rename_new [i]);
      }
    }


    // below this point, need to make sure work==0 before we call cb,
    // since we are running with multiple "threads"
    work++;

    // download new files
    for (unsigned i=0; i<local_add.size (); i++) {
      ude_t *cde = &(upstp->cfs [local_add [i]]);
      bool user_asked_for_this = false;
      if (upstp->filter != "" && upstp->filter == cde->name)
	user_asked_for_this = true;

      str path = make_path (upstp->path, cde->name);
      str cfspath = make_path (upstp->cfspath, cde->name);

      if (!checkonly &&
	  (cde->type != CFSDIR || updatedir || user_asked_for_this)) {
        work++;
	if (update_debug)
	  warn << path << ": download\n";
        checkout_one
	  (cde->fh, path, cfspath, rep_saddr,
	   wrap (mkref (this), &update_object::op_done, upstp, cb));
      }
      else if (checkonly) {
	if (cde->type != CFSDIR)
	  pastwatch_status ("U", path);
	else if (updatedir || user_asked_for_this)
	  pastwatch_status ("D", path);
      }
    }
  
    // do pairwise file update
    for (unsigned i=0; i<upstp->cfs.size (); i++) {

      // We need to make sure renamed files get compared properly
      // if the -n flag is given to pastwatch.

      ude_t *cde = &(upstp->cfs [i]);

      // This says don't compare if the file is new.
      // Maybe we need to compare the file under the old name, to the new named
      // file in the working dir.
      if (cde->status != 'n') {
        unsigned j;
        for (j=0; j<upstp->local.size (); j++) {
          if (cde->fh == upstp->local [j].fh && 
	      !upstp->local [j].compare &&
	      !upstp->local [j].conflict)
	    break;
        }
	if (j < upstp->local.size ()) {
	  ude_t *de = &(upstp->local [j]);
	  de->compare = true;
          work++;
          compare
	    (de, cde,
	     wrap (mkref (this), &update_object::op_done, upstp, cb));
	}
      }
    }

    work--;
    if (work == 0) {
      // set attribute for the whole dir if the dir changed
      if (upstp->filter == "") {
        ptr<pfe> x = pmgr->lookup_in_parent_dir (upstp->path);
        if (x) {
	  if (arg_resetsticky)
	    x->sticky = 0;
	  if (x->sticky == 0) {
	    x->saddr = rep_saddr;
	    if (arg_revision0 || arg_tag0 != "")
	      x->sticky = rep_saddr;
	    else
	      x->sticky = 0;
	  }
	  if (!arg_noupdate) {
	    pmgr->schedule_add (upstp->path, *x);
	    if (arg_command == cmd_merge) {
	      assert (arg_tag0 != "");
              bigint n;
              dearmor64A_bigint (arg_tag0, n);
	      pmgr->schedule_supersedes (upstp->path, n);
	    }
	  }
	}
      }
      update_next_dir (cb);
    }
  }

  void
  cfs_readdir (upstp_t upstp, unsigned idx, cbi cb,
               cfsrcallp c, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << upstp->path << ": cannot read cfs directory(1)\n";
      cb (-1);
      return;
    }
    ude_t *de = &(upstp->cfs [idx]);
    cfsr_readdir_res *res = static_cast<cfsr_readdir_res *> (c->res ());
    de->dirempty = (res->dir->entries == 0L);
    if (work == 0) {
      if (error)
	cb (0);
      else
        compare_directory (upstp, cb);
    }
  }
  
  void
  cfs_getattr (upstp_t upstp, unsigned idx, cbi cb,
               cfsrcallp cc, cfsrw_status status)
  {
    work--;
    ude_t *de = &(upstp->cfs [idx]);
    if (status) {
      warn << de->cfspath << ": cannot get attribute\n";
      cb (-1);
      return;
    }

    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    de->commit = res->attr->commit;
    de->mode_commit = res->attr->mode_commit;
    de->type = res->attr->type;
    de->mode = res->attr->mode;
    de->revision = res->attr->revision;
    de->dirempty = false;

    if (de->type == CFSDIR) {
      work++;
      cfsrcallp c = New refcounted<cfsrcall> 
	(CFSRPROC_READDIR,
	 wrap (mkref (this), &update_object::cfs_readdir, upstp, idx, cb));
      cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
      arg->dir = de->fh;
      arg->cookie = 0;
      arg->count = 2048;
      cfsr_call (rep_saddr, c);
      return;
    }
  
    if (work == 0) {
      if (error)
	cb (0);
      else
	compare_directory (upstp, cb);
    }
  }
  
  void
  local_checklink (upstp_t upstp, unsigned idx, cbi cb,
                   cfsrcallp cc, cfsrw_status status)
  {
    work--;
    ude_t *de = &(upstp->local [idx]);
    if (status) {
      warn << "cannot get attribute of " << armor64A (de->fh) << "\n";
      warn << "it is possible that this file is from another branch!\n";
      cb (-1);
      return;
    }
    else {
      cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
      de->rnlink = res->attr->nlink;
      de->rtype = res->attr->type;

      ude_t cde;
      cde.name = de->name;
      cde.path = de->path;
      cde.cfspath = de->cfspath;
      cde.fh = de->fh;
      cde.dirfh = de->dirfh;
      cde.status = 'A';
      cde.modified = false;
      cde.renamed = false;
      cde.commit = res->attr->commit;
      cde.mode_commit = res->attr->mode_commit;
      cde.type = res->attr->type;
      cde.mode = res->attr->mode;
      cde.revision = res->attr->revision;
      cde.dirempty = false;
      upstp->cfs.push_back (cde);
    }

    if (work == 0) {
      if (error)
	cb (0);
      else
	compare_directory (upstp, cb);
    }
  }

  void
  check_modify (str s, int fd, upstp_t upstp, unsigned idx, cbi cb,
                cfsrw_status status)
  {
    work--;
    close (fd);
    ude_t *de = &(upstp->local [idx]);

    if (status) {
      warn << "cannot read checked out version of the file\n";
      unlink (s);
      cb (-1);
      return;
    }

    int r = isdiff (de->path.cstr (), s.cstr ());
    if (r < 0) {
      warn << de->path << ": cannot compare against checked out version\n";
      unlink (s);
      cb (-1);
      return;
    }
    else if (r > 0){
      de->modified = true;
    }else {
      // warn << "suspected to have been modified, but not true\n";
      de->modified = false;
    }

    if (work == 0) {
      if (error)
	cb (0);
      else
        compare_directory (upstp, cb);
    }
  }

  void
  local_getattr (upstp_t upstp, unsigned idx, cbi cb,
                 cfsrcallp cc, cfsrw_status status)
  {
    work--;
    ude_t *de = &(upstp->local [idx]);
    if (status) {
      warn << "cannot get attribute of " << armor64A (de->fh) << "\n";
      cb (-1);
      return;
    }
    else {
      cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
      de->commit = res->attr->commit;
      de->mode_commit = res->attr->mode_commit;
      de->revision = res->attr->revision;
      // if the entry is faked, update the type; otherwise, we use the
      // type of the file stored on the local filesystem
      if (de->status != ' ' && de->status != '?')
	de->type = res->attr->type;
      else if (de->modified == true && de->mode == CFSREG) {
	// if modified is true, it means that the mtime of the current
	// file in the working directory is different from that
	// recorded in the db file. in this case, try see if there
	// were actual changes made to the local file.

	work++;
	str s = pastwatch_tmpfile ();
        int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
        read_file
          (fd, de->fh, de->saddr,
           wrap (mkref (this), &update_object::check_modify,
	         s, fd, upstp, idx, cb));
	return;
      }
    }

    if (work == 0) {
      if (error)
	cb (0);
      else
        compare_directory (upstp, cb);
    }
  }
  
  void
  get_local (ude_t &de, str name, str path, str cfspath, bigint dirfh)
  {
    de.name = name;
    de.path = path;
    de.cfspath = cfspath;
    de.dirfh = dirfh;
    de.fh = 0;
    de.status = '?';
    de.modified = true;
    de.renamed = false;

    ptr<pfe> v = pmgr->lookup_in_parent_dir (de.path);
    if (v) {
      de.status = ' ';
      de.fh = v->fh;
      de.saddr = v->saddr;
      de.db_mtime = v->mtime;
      if (!arg_resetsticky) {
        de.sticky = v->sticky;
        if (v->sticky != 0 && arg_revision0 == 0 && arg_tag0 == "")
          warn << de.path 
	       << ": comparing working version against sticky version\n";
      }
      else
	de.sticky = 0;
      if (arg_revision0 || arg_tag0 != "")
	// overwrite sticky if -r or -t is specified
	de.sticky = rep_saddr;
      de.newsaddr = rep_saddr;
      if (de.sticky != 0)
	de.newsaddr = de.sticky;
    }

    struct stat sb;
    if (!lstat (de.path, &sb)) {
      // for link, both S_IFLNK and S_IFREG are set, so check for
      // S_IFLNK first
      if (ISLNK (sb))
        de.type = CFSLNK;
      else if (ISREG (sb))
        de.type = CFSREG;
      else if (ISDIR (sb))
        de.type = CFSDIR;
#ifdef SFS_HAVE_STAT_ST_MTIMESPEC
      de.mtime.sec = sb.st_mtimespec.tv_sec;
      de.mtime.nsec = sb.st_mtimespec.tv_nsec;
#else
      de.mtime.sec = sb.st_mtime;
      de.mtime.nsec = 0;
#endif
      de.mode = (unsigned) sb.st_mode;
    }
    else {
      de.mtime.sec = 0;
      de.mtime.nsec = 0;
      de.type = CFSREG;
      // just a guess
      de.mode = 0644;
    }

    if (v && de.mtime.sec == v->mtime.sec && de.mtime.nsec == v->mtime.nsec)
      de.modified = false;
  }

  void
  readdir_reply (upstp_t upstp, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    work--;
    if (status) {
      warn << upstp->path << ": cannot read cfs directory(2)\n";
      cb (-1);
      return;
    }

    cfsr_readdir_res *res = static_cast<cfsr_readdir_res *> (cc->res ());
    rpc_ptr<cfs_rpc_dirent> e = res->dir->entries;
    uint64 cookie = 0;
  
    while (e) {
      if (upstp->filter == "" || upstp->filter == e->name) {
        ude_t de;
        de.name = e->name;
        de.path = make_path (upstp->path, e->name);
        de.cfspath = make_path (upstp->cfspath, e->name);
        de.fh = e->fh;
        de.dirfh = upstp->dirfh;
        de.status = 'n';
        de.modified = false;
        upstp->cfs.push_back (de);

        work++;
        cfsrcallp c = New refcounted<cfsrcall> 
	  (CFSRPROC_GETATTR,
	   wrap (mkref (this), &update_object::cfs_getattr,
	         upstp, upstp->cfs.size ()-1, cb));
        cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
        arg->object = e->fh;
        cfsr_call (rep_saddr, c);
      }
      cookie = e->cookie;
      e = e->nextentry;
    }

    if (!res->dir->eof) {
      work++;
      cfsrcallp c = New refcounted<cfsrcall> 
	(CFSRPROC_READDIR,
	 wrap (mkref (this), &update_object::readdir_reply, upstp, cb));
      cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
      arg->dir = upstp->dirfh;
      arg->cookie = cookie;
      arg->count = 2048;
      cfsr_call (rep_saddr, c);
    }
    else if (work == 0) {
      if (error)
	cb (0);
      else
	compare_directory (upstp, cb);
    }
  }

  void
  update_directory (str d_path, str d_cfspath, str filter, cbi cb)
  {
    ptr<pfe> v = pmgr->lookup (d_path);
    if (!v) {
      warn << "skipping " << d_path << ": not a working directory\n";
      cb (-1);
      return;
    }

    // the basic strategy: create two lists of files, one for local
    // dir, one for remote cfs dir. then we iterate the lists to
    // synchronize the directory.
    
    upstp_t upstp = New refcounted <upst_t>;
    upstp->cfspath = d_cfspath;
    upstp->path = d_path;
    upstp->dirfh = v->fh;
    upstp->filter = filter;

    DIR *dir = opendir (d_path);
    if (dir == 0) {
      perror ("opendir");
      cb (-1);
      return;
    }

    struct dirent *e;
    while ((e = readdir (dir))) {
      bool blah;
      if (check_dirent (upstp->path, e, blah))
	continue;
      bool should_skip = false;
      check_skip (upstp->path, str (e->d_name), should_skip);
      if (should_skip)
	continue;
      if (upstp->filter != "" && upstp->filter != e->d_name)
	continue;
      str fp = make_path (upstp->path, e->d_name);
      str cfsfp = make_path (upstp->cfspath, e->d_name);
      ude_t de;
      get_local (de, e->d_name, fp, cfsfp, upstp->dirfh);
      upstp->local.push_back (de);
      if (e->d_type == DT_DIR && arg_recurse) {
	dirs.push_back (fp);
	filters.push_back ("");
      }
    }
    
    // get scheduled directory operations, augment the local file list
    // to take into account of these operations

    struct dirops dpv;
    if (get_dirops (upstp->path, dpv)) {
      warn << "cannot get list of scheduled directory operations!\n";
      cb (-1);
      return;
    }
    else {
      struct dirops new_dpv;
      new_dpv.clear ();

      // Find new files
      for (unsigned i=0; i<dpv.add.size (); i++) {
        if (upstp->filter != "" && upstp->filter != dpv.add [i]->name) {
	  new_dpv.push_add (dpv, i);
  	  continue;
	}
        str fp = make_path (upstp->path, dpv.add [i]->name);
        bool found = false;
        for (unsigned j=0; j<upstp->local.size () && !found; j++) {
	  if (dpv.add [i]->name == upstp->local [j].name) {
	    found = true;
	    upstp->local [j].status = 'A';
	    upstp->local [j].saddr = 0;
	    if (dpv.add [i]->fh != 0) {
              assert (upstp->local [j].fh == 0);
              upstp->local [j].fh = dpv.add [i]->fh;
	      work++;
              cfsrcallp c = New refcounted<cfsrcall>
		(CFSRPROC_GETATTR,
		 wrap (mkref (this),
		       &update_object::local_checklink, upstp, j, cb));
              cfsr_getattr_arg *arg =
		static_cast<cfsr_getattr_arg *> (c->arg ());
              arg->object = upstp->local [j].fh;
              cfsr_call (rep_saddr, c);
	    }
	  }
        }
        if (!found)
	  warn << fp << " scheduled for add, but disappeared.\n";
        else
	  new_dpv.push_add (dpv, i);
      }

      // Find removed and relinked files
      for (unsigned i=0; i<dpv.remove.size (); i++) {
        if (upstp->filter != "" && upstp->filter != dpv.remove [i]->name) {
	  new_dpv.push_remove (dpv, i);
  	  continue;
	}
        bool found = false;
	unsigned j = 0;
        for (j=0; j<upstp->local.size (); j++)
	  if (dpv.remove [i]->name == upstp->local [j].name) {
	    found = true;
	    break;
	  }
        if (!found) {
	  new_dpv.push_remove (dpv, i);
          str fp = make_path (upstp->path, dpv.remove [i]->name);
          str cfsfp = make_path (upstp->cfspath, dpv.remove [i]->name);
          ude_t de;
          de.name = dpv.remove [i]->name;
          de.path = fp;
          de.cfspath = cfsfp;
          de.dirfh = upstp->dirfh;
          de.fh = dpv.remove [i]->fh;
          de.saddr = dpv.remove [i]->saddr;
          de.status = 'R';
	  de.modified = false;
	  de.mode = 0;
          upstp->local.push_back (de);
	}
	else {
	  new_dpv.push_remove (dpv, i);
	  upstp->local [j].status = 'R';
	  // this tells update code that a removed file re-appeared
	  upstp->local [j].modified = true; 
          upstp->local [j].fh = dpv.remove [i]->fh;
          upstp->local [j].saddr = dpv.remove [i]->saddr;
	}
      }
      if (set_dirops (upstp->path, new_dpv)) {
        warn << "cannot write to directory operations file, abort.\n";
        cb (-1);
        return;
      }
    }

    for (unsigned j=0; j<upstp->local.size (); j++) {
      if (upstp->local [j].fh != 0 && upstp->local [j].saddr != 0) {
        work++;
        cfsrcallp c = New refcounted<cfsrcall> 
	  (CFSRPROC_GETATTR,
	   wrap (mkref (this), &update_object::local_getattr, upstp, j, cb));
        cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
        arg->object = upstp->local [j].fh;
        cfsr_call (upstp->local [j].saddr, c);
      }
    }

    work++;
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_READDIR,
       wrap (mkref (this), &update_object::readdir_reply, upstp, cb));
    cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
    arg->dir = upstp->dirfh;
    arg->cookie = 0;
    arg->count = 2048;
    cfsr_call (rep_saddr, c);
  }

  void
  start_update (str path) 
  {
    struct stat sb;
    if (!lstat (path, &sb) && (ISDIR(sb))) {
      // if <path> is a directory, then update the whole directory.
      dirs.push_back (path);
      filters.push_back ("");
      update_next_dir (wrap (mkref (this), &update_object::update_done));
    }
    else {
      // Else, just update the file <n>
      str n;
      str d = get_parent_dir (path, n);
      dirs.push_back (d);
      filters.push_back (n);
      update_next_dir (wrap (mkref (this), &update_object::update_done));
    }
  }

  void
  findver_cb (str path, cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot find revision " << arg_revision0 << ".  Maybe the branch is incorrect.\n";
      final_cb (-1);
      return;
    }
    // found the old revision that we're updating to.
    cfsr_findver_res *res = static_cast<cfsr_findver_res *> (c->res ());
    rep_saddr = res->saddr;
    start_update (path);
  }

  void
  start (str path)
  {
    ptr<pfe> v = pmgr->lookup (path);

    // if -r is specified, search for the old repository snapshot
    if (arg_revision0) {
      if (!v) {
        warn << "no information for " << path << ", not checked-in?\n";
        final_cb (0);
        return;
      }
      cfsrcallp c = New refcounted<cfsrcall>
	(CFSRPROC_FINDVER,
	 wrap (mkref (this), &update_object::findver_cb, path));
      cfsr_findver_arg *a = static_cast<cfsr_findver_arg *> (c->arg ());
      a->object = v->fh;
      a->revision = arg_revision0;
      cfsr_call (rep_saddr, c);
    }
    else {
      if (arg_tag0 == "" && !arg_resetsticky && v && v->sticky != 0)
	rep_saddr = v->sticky;
      start_update (path);
    }
  }

public:
  update_object (bigint root, str path, cbi cb,
                 bool checkonly, bool updatedir)
    : work (0), error (0), final_cb (cb),
      checkonly (checkonly), updatedir (updatedir)
  {
    if (arg_tag0 != "") {
      bigint n;
      dearmor64A_bigint (arg_tag0, n);
      ptr<btree_srv> srv = _cfsrw->getreplica ()->btree ();
      btree_node x;
      srv->lookup_by_commit (n, x);
      rep_saddr = x.commit.vptr;
    }
    else
      rep_saddr = my_branch_head_snapshot (_cfsrw->getreplica ());

    rep_root = root;
    requires_update = false;
    has_conflict = false;
    dirs.clear ();
    filters.clear ();
    to_commit.clear ();
    to_setattr.clear ();
    to_remove.clear ();
    to_add.clear ();
    delaycb (0, wrap (mkref (this), &update_object::start, path));
  }
};

void
update (bigint root, str path, cbi cb)
{
  ptr<update_object> a = New refcounted<update_object>
    (root, path, cb, arg_checkonly, arg_updatedir);
}

ptr<update_object_spec>
update_object_spec::alloc (bigint root, str path, cbi cb,
                           bool checkonly, bool updatedir)
{
  ptr<update_object> a = New refcounted<update_object>
    (root, path, cb, checkonly, updatedir);
  return a;
}

