/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "parseopt.h"
#include "rxx.h"
#include "ns.h"
#include "logw.h"
#include "record.h"

static bigint head;
static bigint view;
static bigint hash;
static ptr<sfspriv> sk;
ptr<ns> nsobj;

int ns_debug = 0;

void
store_block_cb (ns_stat status, bigint key)
{
  assert (key == head);
  if (status != NS_OK) {
    warn << "cannot store log head " << armor64A (head) << "\n";
    exit (1);
  }
  exit (0);
}

static void
load_head (bigint n, ptr<log_pointer> r, long v, ns_stat status)
{
  if (!r) {
    warn << "cannot load log head " << armor64A (head) << "\n";
    exit (1);
  }

  warn << "set last commit to " << armor64A (n) << "\n";
  r->set_commit (n);
  nsobj->insert_pk (sk, r->serialize (), wrap (store_block_cb));
}

static void
load_commit (bigint n, ptr<commit_record> r, ns_stat status)
{
  if (!r) {
    warn << "cannot load commit record " << armor64A (n) << "\n";
    exit (1);
  }

  if (r->head () == head)
    nsobj->retrieve_lptr (head, wrap (load_head, n));
  else {
    warn << armor64A (n) << ": not your commit record, skip\n";
    if (r->depends ().rptr != 0) {
      n = r->depends ().rptr;
      nsobj->retrieve_commit (n, wrap (load_commit, n));
    }
  }
}

void
start ()
{
  nsobj->retrieve_commit (hash, wrap (load_commit, hash));
}

static void
usage ()
{
  warnx << "usage: " << progname << " -s <keyfile> -p <project> hash\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  char *sk_file = 0;
  str vs;
  
  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "s:p:")) != -1)
    switch (ch) {
    case 's':
      sk_file = optarg;
      break;
    case 'p':
      vs = str (optarg, strlen (optarg));
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if (argc != 1 || !sk_file )
    usage ();

  str x = str (argv [0], strlen (argv [0]));
  if (!dearmor64A_bigint (x, hash))
    usage ();
  
  if (!dearmor64A_bigint (vs, view))
    usage ();

  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    fatal ("cannot read key file!\n");
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    exit (1);
  }

  head = ns::pkid (sk);
  nsobj = ns::alloc (view);
  nsobj->connect (wrap (start));
  amain();
}

